import pandas as pd
import pysam
import copy
import subprocess
import json

from celescope.tools import analysis_wrapper
from collections import defaultdict
from celescope.tools import utils
from celescope.tools.capture.threshold import Auto
from celescope.tools.step import Step, s_common
from celescope.flv_trust4.__init__ import CHAIN, PAIRED_CHAIN, TOOLS_DIR
from celescope.tools.emptydrop_cr import get_plot_elements


ASSEMBLE_SUFFIX = 'assembled_reads.fa'
ANNOTATE_SUFFIX = 'annotate.fa'
TRUST_REPORT_SUFFIX = 'filter_report.tsv'
BARCODE_REPORT_SUFFIX = 'barcode_report.tsv'
BARCODE_REPORT_FILTER_SUFFIX = 'barcode_filter_report.tsv'


def target_cell_calling(df_UMI_sum, expected_target_cell_num=3000, target_barcodes=None, weight=6, coef=5, 
    percentile=85, umi_col='umis'):
    """
    Args:
        df_UMI_sum: A dataframe with columns highest umi's contig and UMI.
    
    Returns:
        target_contigs_id: list
    >>> df_UMI_sum = pd.DataFrame({"contig_id": ["A", "B", "C", "D", "E"], "UMI": [1, 2, 1, 30, 40]})
    >>> target_contigs_id = target_cell_calling(df_UMI_sum, expected_target_cell_num=5, percentile=80, coef=5, target_barcodes=["A", "C"])
    >>> target_contigs_id == {'A_1', 'C_1', 'D_1', 'E_1'}
    True
    """
    if target_barcodes != None:
        target_barcodes = {i for i in target_barcodes}
    umi_threshold = Auto(list(df_UMI_sum[umi_col]), expected_cell_num=expected_target_cell_num, coef=coef, percentile=percentile).run()

    # avoid change the original dataframe
    df_temp = df_UMI_sum.copy()
    if target_barcodes:
        df_temp[umi_col] = df_temp.apply(
            lambda row:  row[umi_col] * weight if row['barcode'] in target_barcodes else row[umi_col], axis=1)
             
    target_contigs = set(df_temp.loc[df_temp[umi_col] >= umi_threshold].contig_id)

    return target_contigs


class Summarize(Step):
    """
    ## Features
    - CDR3 filtering: contain stop condon, length <=5, etc..

    - If barcode A's two chains CDR3s are identical to another barcode B, and A's chain abundance is significantly lower than B's, filter A.

    - If `--target_cell_barcode` is provided, the UMI counts of all contigs originated from target cells are multiplied by a weight(default: 6.0) to better distinguish signal from background noise. `--target_cell_barcode` comes from the cell type annotation results of the RNA library.

    - Cell-calling is similar to the rna cell-calling algorithm.

    ## Output
    - `04.summarize/clonetypes.tsv` High-level description for each clonotype.

    - `04.summarize/{sample}_all_contig.csv` High-level and detailed annotation for each contig.

    - `04.summarize/{sample}_all_contig.fasta` All assembled contig sequences.

    - `04.summarize/{sample}_filtered_contig.csv` High-level annotations of each cellular contig after filter. This is a subset of {sample}_all_contig.csv.
    
    - `04.summarize/{sample}_filtered_contig.fasta` Assembled contig sequences after filter.
    """
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.ref = args.ref
        self.fq2 = args.fq2
        self.diffuseFrac = args.diffuseFrac
        self.assembled_fa = f'{args.assemble_out}/{self.sample}_{ASSEMBLE_SUFFIX}'
        self.trust_report = f'{args.assemble_out}/{self.sample}_{TRUST_REPORT_SUFFIX}'
        self.annot = f'{args.assemble_out}/{self.sample}_{ANNOTATE_SUFFIX}'

        self.match_dict = utils.parse_match_dir(args.match_dir)
        self.chains, self.paired_groups = self._parse_seqtype(self.seqtype)

        # if --diffuseFrac provided
        if self.diffuseFrac:
            self.barcode_report = f'{args.assemble_out}/{self.sample}_{BARCODE_REPORT_FILTER_SUFFIX}'
        else:
            self.barcode_report = f'{args.assemble_out}/{self.sample}_{BARCODE_REPORT_SUFFIX}'
        
        self.coef = int(args.coef)
        self.target_weight = args.target_weight

        """
        There are three method. 
        1. --expected_target_cell_num, Expected assembled T or B cell number. eg 3000
        2. --target_cell_barcode Auto, Based on auto assigned cell type annotation results of the RNA library.
        3. --target_cell_barcode Absolute path of a plain text file with one barcode per line.(Recommend, need manual-assign first)
        """
        if not args.target_cell_barcode:
            self.target_barcodes = None
            self.expected_target_cell_num = args.expected_target_cell_num
        elif args.target_cell_barcode == 'Auto':
            self.target_barcodes, self.expected_target_cell_num = self._auto_assign()
        else:
            self.target_barcodes, self.expected_target_cell_num = utils.read_one_col(args.target_cell_barcode)
    
    @staticmethod
    def _parse_seqtype(seqtype):
        """Parse BCR or TCR

        :param seqtype
        :return CHAIN[seqtype]: 'TCR': ['TRA', 'TRB'], 'BCR': ['IGH', 'IGL', 'IGK']
        	    PAIRED_CHAIN[seqtype]: ['TRA_TRB'], 'BCR': ['IGH_IGL', 'IGH_IGK']
        """
        return CHAIN[seqtype], PAIRED_CHAIN[seqtype]
    
    @staticmethod
    @utils.add_log
    def convert_barcode_report(barcode_report, outdir):
        cmd = (
            f'perl {TOOLS_DIR}/trust-barcoderep-to-10X.pl '
            f'{barcode_report} '
            f'{outdir} '
        )
        Summarize.convert_barcode_report.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)
    
    def add_cell_num_metric(self, df, name):
        """
        add cell number after each filtering step to .metrics.json
        do not show in HTML report
        """
        cell_num = len(set(df[df['productive']==True].barcode))
        self.add_metric(name, cell_num, show=False)
        return cell_num
    
    @utils.add_log
    def _auto_assign(self):
        """
        --target_cell_barcode Auto
        Auto assign cell type of the RNA library based on marker gene file.
        target_cell_barcode requires >=2 identified marker genes of T/B cells in top10 marker genes for each cluster.
        Return target_barcodes, expected_target_cell_num.
        """
        species, cell_type = Summarize.get_cell_species(self.ref, self.seqtype)

        with open(f'{TOOLS_DIR}/Immune_marker.json', 'r') as f:
            marker_dict = json.load(f)

        report_runner = analysis_wrapper.Report_runner(self.args)
        df_tsne, df_marker = report_runner.get_df()
        
        bc_cluster_dict = df_tsne.reset_index().groupby('cluster')['barcode'].apply(lambda x: x.tolist()).to_dict()
        df_marker_top5 = df_marker.groupby('cluster').head(5)
        df_marker_top5['celltype'] = None

        target_genes = set(df_marker_top5.gene)
        for gene in target_genes:
            if gene in marker_dict[species][cell_type]:
                df_marker_top5.loc[df_marker_top5['gene']==gene, 'celltype'] = cell_type

        target_clusters = df_marker_top5[df_marker_top5['celltype']==cell_type].cluster.tolist()
        target_clusters = [int(x.split(' ')[-1]) for x in target_clusters]
        target_clusters = set([x for x in target_clusters if target_clusters.count(x) > 1])
        
        if not target_clusters:
            return None, self.args.expected_target_cell_num

        target_barcodes = []
        for cluster in target_clusters:
            target_barcodes += bc_cluster_dict[cluster]

        return target_barcodes, len(target_barcodes)

    @staticmethod
    def get_cell_species(ref, seqtype):

        if ref == 'GRCm38':
            species = 'mouse'
        else:
            species = 'human'

        if seqtype =='BCR':
            target_cell_type = 'B_cells'
        else:
            target_cell_type = 'T_cells'

        return species, target_cell_type

    @utils.add_log
    def parse_contig_file(self):
        """
        Convert to all_contig_annotation file in 10X format.
        Generate all_contig_fasta file.
        """
        self.convert_barcode_report(self.barcode_report, outdir=f'{self.outdir}/{self.sample}')

        if self.seqtype == 'BCR':
            df = pd.read_csv(f'{self.outdir}/{self.sample}_b.csv')
        else:
            df = pd.read_csv(f'{self.outdir}/{self.sample}_t.csv')

        df['productive'] = df['full_length']
        contig_set = set(df.contig_id)

        # generate all contig fasta file
        # add length of each contig. 
        len_dict = dict()
        all_fa = open(f'{self.outdir}/{self.sample}_all_contig.fasta','w')
        with pysam.FastxFile(self.annot) as fa:
            for read in fa:
                len_dict[read.name] = read.comment.split(' ')[0]
                if read.name in contig_set:
                    sequence = read.sequence
                    all_fa.write('>' + read.name + '\n' + sequence + '\n')    
        all_fa.close()
        df['length'] = df['contig_id'].apply(len_dict.get)
        
        return df

    @utils.add_log
    def cell_calling(self, df):
        """
        Common filtering based on CDR3:
        Filter nonfunctional CDR3(shown 'out_of_frame' in cdr3 report), or CDR3 sequences containing "N" in the nucleotide sequence.
        Keep CDR3aa start with C.
        Keep CDR3aa length >= 5.
        Keep no stop codon in CDR3aa.
        Filter low abundance contigs based on a umi cut-off. 

        DiffuseFrac filtering(option --diffuseFrac needed):
        If cell A's two chains CDR3s are identical to another cell B, 
        and A's chain abundance is significantly lower than B's (--diffuseFrac), filter A.

        Target cell barcodes filtering(option, --target_cell_barcode needed):
        Filter low abundance contigs based on a umi cut-off.
        The umi counts of all contigs originated from B cells are multiplied by a weight to 
        better distinguish signal from background noise.
        """
        df.sort_values(by='umis', ascending=False, inplace=True)
        if self.seqtype == 'BCR':
            df_chain_heavy = df[df['chain']=='IGH']
            df_chain_light = df[(df['chain']=='IGK') | (df['chain']=='IGL')]
        else:
            df_chain_heavy = df[df['chain'] == 'TRA']
            df_chain_light = df[df['chain'] == 'TRB']
        df_chain_heavy = df_chain_heavy.drop_duplicates(['barcode'])
        df_chain_light = df_chain_light.drop_duplicates(['barcode'])
        df_for_clono = pd.concat([df_chain_heavy, df_chain_light], ignore_index=True)
        
        self.add_cell_num_metric(df_for_clono, 'Total Assembled Cell Number')
        
        # Common filtering
        trust_report = pd.read_csv(self.trust_report, sep='\t')
        correct_cdr3 = set(df_for_clono.cdr3).intersection(set(trust_report.CDR3aa))
        correct_cdr3 = [i for i in correct_cdr3 if i.startswith('C')]
        correct_cdr3 = [i for i in correct_cdr3 if len(i)>=5]
        correct_cdr3 = [i for i in correct_cdr3 if 'UAG' or 'UAA' or 'UGA' not in i]
        df_for_clono = df_for_clono[df_for_clono['cdr3'].isin(correct_cdr3)]

        self.add_cell_num_metric(df_for_clono, 'Cell Number after CDR3 filtering')
        
        # Filter low abundance contigs based on a umi cut-off
        if self.seqtype == 'BCR':
            df_chain_heavy = df_for_clono[df_for_clono['chain']=='IGH']
            df_chain_light = df_for_clono[(df_for_clono['chain']=='IGK') | (df_for_clono['chain']=='IGL')]
        else:
            df_chain_heavy = df_for_clono[df_for_clono['chain'] == 'TRA']
            df_chain_light = df_for_clono[df_for_clono['chain'] == 'TRB']

        filtered_congtigs_id = set()
        for _df in [df_chain_heavy, df_chain_light]:
            target_contigs = target_cell_calling(
            _df, 
            expected_target_cell_num=self.expected_target_cell_num, 
            target_barcodes=self.target_barcodes,
            weight = self.target_weight,
            coef = self.coef
            )
            filtered_congtigs_id = filtered_congtigs_id | target_contigs       
        
        df_for_clono = df_for_clono[df_for_clono.contig_id.isin(filtered_congtigs_id)]
        self.add_cell_num_metric(df_for_clono, 'Cell Number after UMI filtering')
        
        df_for_clono_pro = df_for_clono[df_for_clono['productive']==True]
        cell_barcodes, filtered_contig = set(df_for_clono_pro['barcode']), set(df_for_clono_pro['contig_id'])


        return df_for_clono, cell_barcodes, filtered_contig

    @utils.add_log
    def filter_fasta(self, cell_barcodes):
        """Filter all contig fasta file by barcodes which are identified to be cell.

        :param cell_barcodes: all barcodes identified to be cell.
        """
        all_contig_fasta = f'{self.outdir}/{self.sample}_all_contig.fasta'
        filter_contig_fasta = f'{self.outdir}/{self.sample}_filtered_contig.fasta'

        filter_contig_fasta = open(filter_contig_fasta,'w')
        with pysam.FastxFile(all_contig_fasta) as fa:
            for read in fa:
                name = read.name
                cb = name.split(':')[0]
                sequence = read.sequence
                if cb in cell_barcodes:
                    filter_contig_fasta.write('>' + name + '\n' + sequence + '\n')

        filter_contig_fasta.close()

    @utils.add_log
    def parse_clonotypes(self, df, df_for_clono, cell_barcodes, filtered_contig):
        """Parse clonotypes from CDR3 and manually add clonotype id for each contig.

        :param df: original contig file.
        :param df_for_clono: contig info after filter.
        :param cell_barcodes: all barcodes identified to be cell.
        :return df_filter_contig: filtered contigs by cell barcodes.
        """
        df_for_clono_pro = df_for_clono[df_for_clono['productive']==True].copy()
        df_for_clono_pro['chain_cdr3aa'] = df_for_clono_pro.loc[:, ['chain', 'cdr3']].apply(':'.join, axis=1)
        df_for_clono_pro['chain_cdr3nt'] = df_for_clono_pro.loc[:,['chain', 'cdr3_nt']].apply(':'.join, axis=1)

        cbs = set(df_for_clono_pro['barcode'])
        clonotypes = open(f'{self.outdir}/clonotypes.csv', 'w')
        clonotypes.write('barcode\tcdr3s_aa\tcdr3s_nt\n')
        for cb in cbs:
            temp = df_for_clono_pro[df_for_clono_pro['barcode']==cb]
            temp = temp.sort_values(by='chain', ascending=True)
            aa_chain = ';'.join(list(temp['chain_cdr3aa']))
            nt_chain = ';'.join(list(temp['chain_cdr3nt']))
            clonotypes.write(f'{cb}\t{aa_chain}\t{nt_chain}\n')
        clonotypes.close() 

        df_clonotypes = pd.read_csv(f'{self.outdir}/clonotypes.csv', sep='\t', index_col=None)
        contig_with_clonotype = copy.deepcopy(df_clonotypes)
        df_dict = df_clonotypes[["cdr3s_nt", "cdr3s_aa"]].set_index("cdr3s_nt").to_dict(orient='dict')['cdr3s_aa']
        df_clonotypes = df_clonotypes.groupby('cdr3s_nt', as_index=False).agg({'barcode': 'count'})
        df_clonotypes.rename(columns={'barcode': 'frequency'}, inplace=True)
        sum_f = df_clonotypes['frequency'].sum()

        df_clonotypes['proportion'] = df_clonotypes['frequency'].apply(lambda x: x/sum_f)
        df_clonotypes.sort_values(by='frequency', ascending=False, inplace=True)
        df_clonotypes['clonotype_id'] = [f'clonotype{i}' for i in range(1, df_clonotypes.shape[0]+1)]
        df_clonotypes['cdr3s_aa'] = df_clonotypes['cdr3s_nt'].apply(lambda x:df_dict[x])
        df_clonotypes = df_clonotypes.reindex(columns=['clonotype_id', 'frequency', 'proportion', 'cdr3s_aa', 'cdr3s_nt'])
        df_clonotypes.to_csv(f'{self.outdir}/clonotypes.csv', sep=',', index=False) 
        used_for_merge = df_clonotypes[['cdr3s_nt','clonotype_id']]

        df_merge = pd.merge(used_for_merge, contig_with_clonotype, on='cdr3s_nt', how='outer')
        df_merge = df_merge[['barcode', 'clonotype_id']]
        df_all_contig = pd.merge(df_merge, df, on='barcode',how='outer')
        df_all_contig.fillna('',inplace = True)
        df_all_contig = df_all_contig[['barcode', 'is_cell', 'contig_id', 'high_confidence', 'length', 'chain', 'v_gene', 'd_gene', 'j_gene', 'c_gene', 'full_length', 'productive', 'cdr3', 'cdr3_nt', 'reads', 'umis', 'clonotype_id']]
        df_filter_contig = df_all_contig[df_all_contig['barcode'].isin(cell_barcodes)]
        for _df in [df_all_contig, df_filter_contig]:
            _df.loc[~_df.contig_id.isin(filtered_contig), 'clonotype_id'] = ''

        df_all_contig.to_csv(f'{self.outdir}/{self.sample}_all_contig.csv', sep=',', index=False)
        df_filter_contig.to_csv(f'{self.outdir}/{self.sample}_filtered_contig.csv', sep=',', index=False)

    @utils.add_log
    def gen_summary(self, df_for_clono):
        """ Generate metrics in html 
        """
        df_for_clono_pro = df_for_clono[df_for_clono['productive']==True]
        cell_barcodes = set(df_for_clono_pro['barcode'])
        total_cells =len(cell_barcodes)

        read_count = 0
        read_count_all = 0
        umi_dict = defaultdict(set)
        umi_count = defaultdict()
        with pysam.FastxFile(self.fq2) as fq:
            for read in fq:
                read_count_all+=1
                cb = read.name.split(':')[0]
                umi = read.name.split(':')[1]
                umi_dict[cb].add(umi)
                if cb in cell_barcodes:
                    read_count+=1
        for cb in umi_dict:
            umi_count[cb] = len(umi_dict[cb])
        df_umi = pd.DataFrame.from_dict(umi_count, orient='index', columns=['UMI'])
        df_umi['barcode'] = df_umi.index
        df_umi = df_umi.reset_index(drop=True)
        df_umi = df_umi.reindex(columns=['barcode', 'UMI'])
        df_umi = df_umi.sort_values(by='UMI', ascending=False)
        df_umi['mark'] = df_umi['barcode'].apply(lambda x: 'CB' if x in cell_barcodes else 'UB')
        df_umi.to_csv(f'{self.outdir}/count.txt', sep='\t', index=False)
        self.add_data(chart=get_plot_elements.plot_barcode_rank(f'{self.outdir}/count.txt'))

        self.add_metric(
            name = 'Estimated Number of Cells',
            value = total_cells,
            help_info = "Number of cells which contain at least one chain (for TCR: TRA or TRB, for BCR: IGH, IGL or IGK)"            
        )

        self.add_metric(
            name = 'Mean Read Pairs per Cell',
            value = int(read_count/total_cells),
            help_info = 'Number of input reads divided by the estimated number of cells'
        )

        used_read = 0
        with pysam.FastxFile(self.assembled_fa) as fa:
            for read in fa:
                bc = read.name.split(':')[0]
                if bc in cell_barcodes:
                    used_read += 1
        self.add_metric(
            name = 'Mean Used Read Pairs per Cell',
            value = int(used_read/total_cells), 
            help_info = "Mean number of reads used in assembly per cell-associated barcode"
        )
        self.add_metric(
            name = 'Fraction of Reads in Cells',
            value = used_read,
            total = read_count_all,
            help_info = 'Number of reads with cell-associated barcodes divided by the number of reads with valid barcodes'
        )
        
        for c in self.chains:
            temp_df = df_for_clono_pro[df_for_clono_pro['chain']==c]

            try:
                median_umi = int(temp_df['umis'].median())
            except ValueError:
                # ValueError: cannot convert float NaN to integer
                median_umi = 0

            self.add_metric(
                name = f'Median {c} UMIs per Cell',
                value = median_umi
            )

    def run(self):
        original_df = self.parse_contig_file()
        df_for_clono, cell_barcodes, filtered_contig = self.cell_calling(original_df)
        self.filter_fasta(cell_barcodes)
        self.parse_clonotypes(original_df, df_for_clono, cell_barcodes, filtered_contig)
        self.gen_summary(df_for_clono)


@utils.add_log
def summarize(args):
    with Summarize(args, display_title="Cells") as runner:
        runner.run()


def get_opts_summarize(parser, sub_program):
    parser.add_argument('--seqtype', help='TCR or BCR', choices=['TCR', 'BCR'], required=True)
    parser.add_argument('--ref', help='reference name', choices=["hg19", "hg38", "GRCm38", "other"], required=True)
    parser.add_argument('--coef', help='coef for auto filter', default=5)
    parser.add_argument(
        '--diffuseFrac', 
        help="If cell A's two chains CDR3s are identical to another cell B, and A's chain abundance is significantly lower than B's, filter A.",
        action='store_true')
    parser.add_argument(
        "--expected_target_cell_num", 
        help="Expected T or B cell number. If `--target_cell_barcode` is provided, this argument is ignored.", 
        type=int,
        default=3000,
    )
    parser.add_argument(
        '--target_cell_barcode', 
        help="Barcode of target cells. Auto or path of plain text file with one barcode per line",
        default=None)
    parser.add_argument(
        "--target_weight", 
        help="UMIs of the target cells are multiplied by this factor. Only used when `--target_cell_barcode` is provided.", 
        type=float,
        default=6.0,
    )
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--fq2', help='Barcode R2 reads.', required=True)
        parser.add_argument('--assemble_out', help='Result of  assemble dirctory.', required=True)
        parser.add_argument('--match_dir', help='Match scRNA-seq directory.', required=True)



import pysam

import pandas as pd

from celescope.tools.capture.count_bam import Count_bam, get_opts_count_bam
from celescope.fusion.mkref import Mkref_fusion
from celescope.tools import utils



class Count_fusion(Count_bam):
    """
    ## Features
    - Count the number of reads and umis that 
        1. originate from cell barcodes;
        2. align to the fusion site and include flanking sequences of a certain length(default 20bp) on both sides of the fusion site.
    """
    def __init__(self, args, display_title='Count'):
        super().__init__(args, display_title)

        self.flanking_base = int(args.flanking_base)
        fusion_pos_file = Mkref_fusion.get_config(args.fusion_genomeDir)['files']['fusion_pos']
        self.pos_dict = self.read_pos_file(fusion_pos_file)
        self.posSorted_bam = f'{self.out_prefix}_posSorted.bam'
        self.fusion_bam = f'{self.out_prefix}_raw_fusion.bam'
 
    @staticmethod
    def read_pos_file(fusion_pos_file):
        """
        pos_dic
            key: sequence name in fusion.fasta
            value: end postion of the first gene(1-based).
        """
        pos_dict = {}
        df = pd.read_csv(fusion_pos_file, sep="\t")
        for name, pos in zip(df.iloc[:, 0], df.iloc[:, 1]):
            pos_dict[name] = pos
        return pos_dict

    def process_bam(self):
        """
        find valid fusion reads
            1. flank the fusion position
            2. match barcode
        """

        with pysam.AlignmentFile(self.posSorted_bam, "rb") as bam:
            header = bam.header
            with pysam.AlignmentFile(self.fusion_bam, "wb", header=header) as fusion_bam:
                for ref in self.pos_dict:
                    pos = self.pos_dict[ref]
                    left = pos - self.flanking_base
                    right = pos + self.flanking_base
                    for read in bam.fetch(
                        reference=ref,
                        start=left,
                        end=right,
                    ):
                        left_bases = read.get_overlap(left, pos)
                        right_bases = read.get_overlap(pos, right)
                        if left_bases < self.flanking_base or right_bases < self.flanking_base:
                            continue
                        attr = read.query_name.split(':')
                        barcode = attr[0]
                        umi = attr[1]
                        if barcode in self.match_barcode:
                            fusion_bam.write(read)
                            self.count_dict[barcode][ref][umi] += 1

    def run(self):
        utils.sort_bam(
            self.args.capture_bam,
            self.posSorted_bam,
            threads=self.thread,
        )
        utils.index_bam(self.posSorted_bam)
        super().run()

def count_fusion(args):
    with Count_fusion(args) as runner:
        runner.run()


def get_opts_count_fusion(parser, sub_program):
    parser.add_argument('--fusion_genomeDir', help='Fusion genome directory.', required=True)
    parser.add_argument(
        "--flanking_base",
        help="Number of bases flanking the fusion position.",
        default=5)
    get_opts_count_bam(parser, sub_program)





from celescope.tools.make_ref import MakeRef_STAR

class Mkref_fusion(MakeRef_STAR):
    """
    ## Features
    - Create a fusion genome directory.

    ## Output

    - STAR genome index files
    - Genome config file

    ## Usage
    ```
    celescope fusion mkref \\
    --genome_name {genome_name} \\
    --fasta fusion.fasta \\
    --fusion_pos fusion_pos.txt \\
    ```
    """
    def __init__(self, genome_type, args):
        super().__init__(genome_type, args)
        self.files['fusion_pos'] = args.fusion_pos


def mkref(args):
    genome_type = 'fusion'
    with Mkref_fusion(genome_type, args) as runner:
        runner.run()


def get_opts_mkref(parser, sub_program):
    MakeRef_STAR.opts(parser, sub_program)
    if sub_program:
        parser.add_argument(
            "--fusion_pos",
            help="""
fusion position file. A two column tab-delimited text file with header.
"pos" is the end postion of the first gene(1-based).
e.g.  
name\tpos  
PML_3\t183  
PML_4\t254  
PML_5\t326  
PML_6\t204   
""",
            required=True,)




from celescope.tools import utils
from celescope.tools.star_mixin import Star_mixin, get_opts_star_mixin


class Star_virus(Star_mixin):
    """
    ## Features
    - Map reads to the viral genome using STAR.

    ## Output
    - `{sample}_virus_Aligned.sortedByCoord.out.bam` : Aligned BAM sorted by coordinate.
    """

    def __init__(self, args, display_title=None):
        # before init
        args.genomeDir = args.virus_genomeDir
        super().__init__(args, add_prefix='virus', display_title=display_title)


@utils.add_log
def star_virus(args):
    with Star_virus(args, display_title='Mapping') as runner:
        runner.run()


def get_opts_star_virus(parser, sub_program):
    get_opts_star_mixin(parser, sub_program)
    parser.add_argument('--virus_genomeDir', help='Virus genome directory.', required=True)


import argparse
import sys
import subprocess

from celescope.tools.step import Step
from celescope.tools.barcode import get_opts_barcode
from celescope.tools.cutadapt import get_opts_cutadapt, get_cutadapt_cmd, Cutadapt as Ct
from celescope.rna.star import get_opts_star
from celescope.tools.star_mixin import get_star_cmd, Star_mixin
from celescope.tools import utils


def args_to_list(args):
    s = []
    for arg, v in vars(args).items():
        if v not in (False, None, '') and arg != 'func':
            s.append(f'--{arg}')
            if v != True:
                if 'param' in arg:
                    s.append(f'"{v}"')
                else:
                    s.append(f'{v}')
    return s


def get_step_args_str(args, get_opts_func):
    args_str = ''
    parser = argparse.ArgumentParser()
    get_opts_func(parser, sub_program=True)
    # [known_args_namespace, not_known_args_list]
    known_args = parser.parse_known_args(args_to_list(args))[0]
    args_str = ' '.join(args_to_list(known_args))
    return args_str


class Prep(Step):
    def __init__(self, args):
        super().__init__(args)

    @utils.add_log
    def run(self):
        bc2 = f'{self.out_prefix}_bc2.FIFO'
        ct2 = f'{self.out_prefix}_ct2.FIFO'
        bc_args_str = get_step_args_str(self.args, get_opts_barcode)
        ct_cmd = get_cutadapt_cmd(self.args, bc2, ct2)
        star_cmd = get_star_cmd(self.args, ct2, self.out_prefix)
        cmd = (
            f'mkfifo {bc2} {ct2} \n'
            f'celescope rna barcode {bc_args_str} --stdout >> {bc2} & \n'
            f'{ct_cmd} & \n'
            f'{star_cmd} \n'
            f'rm {bc2} {ct2} \n'
        )
        sys.stderr.write(cmd + '\n')
        subprocess.check_call(cmd, shell=True)


class Cutadapt(Ct):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

    def run(self):
        self.add_cutadapt_metrics()

class Star(Star_mixin):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

    def run(self):
        self.sort_bam()
        self.index_bam()
        self.remove_unsort_bam()
        self.add_star_metrics()


@utils.add_log
def prep(args):
    runner = Prep(args)
    runner.run()

    with Cutadapt(args, display_title="Trimming") as runner:
        runner.run()

    with Star(args, display_title="Mapping") as runner:
        runner.run()

def get_opts_prep(parser, sub_program):
    get_opts_barcode(parser, sub_program=sub_program)
    get_opts_cutadapt(parser, sub_program=False)
    get_opts_star(parser, sub_program=False)



import subprocess

import pandas as pd

from celescope.tools import utils
from celescope.tools.make_ref import MakeRef
from celescope.tools.step import s_common
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step
from celescope.tools.__init__ import STAR_BAM_SUFFIX


@utils.add_log
def get_star_cmd(args, input_file, output_prefix):
    """
    output sam format to improve speed
    """
    cmd = (
        f'STAR '
        f'--runThreadN {args.thread} '
        f'--genomeDir {args.genomeDir} '
        f'--outSAMmultNmax 1 '
        f'--outFilterMultimapNmax {args.outFilterMultimapNmax} '
        f'--outSAMtype BAM Unsorted '
        f'--outFilterMatchNmin {args.outFilterMatchNmin} '
        f'{args.STAR_param} '
        f'--readFilesIn {input_file} '
        f'--outFileNamePrefix {output_prefix}_ '
    )
    get_star_cmd.logger.info(cmd)
    return cmd


def get_star_log(logPath):
    df = pd.read_csv(logPath, sep='\t',header=None, names=['name','value'])
    log_dict = {}
    for t in df.itertuples():
        name = t.name.strip('|').strip()
        log_dict[name] = t.value

    return log_dict

class Star_mixin(Step):
    """
    Mixin class for STAR
    """

    def __init__(self, args, add_prefix=None, display_title=None):
        super().__init__(args, display_title)

        # parse
        self.genome_name = MakeRef.get_config(args.genomeDir)['meta']['genome_name']
        self.stat_prefix = 'Reads'
        if getattr(args, "consensus_fq", False):
            self.stat_prefix = 'UMIs'

        # out
        if add_prefix:
            self.out_prefix += f'_{add_prefix}'
        self.logPath = f'{self.out_prefix}_Log.final.out'
        self.STAR_bam = f'{self.out_prefix}_{STAR_BAM_SUFFIX}'

    @utils.add_log
    def STAR(self):
        input_file = self.args.fq
        output_prefix = self.out_prefix
        cmd = get_star_cmd(self.args, input_file, output_prefix)

        if self.args.out_unmapped:
            cmd += '--outReadsUnmapped Fastx'
        if self.args.fq[-3:] == ".gz":
            cmd += '--readFilesCommand zcat'
        if self.args.STAR_param:
            cmd += (" " + self.args.STAR_param)
        self.STAR.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        self.STAR()
        self.add_star_metrics()

    def add_star_metrics(self):
        """
        step metrics
        """
        log_dict = get_star_log(self.logPath)

        unique_reads = int(log_dict['Uniquely mapped reads number'])
        multiple_loci_reads = int(log_dict['Number of reads mapped to multiple loci'])
        too_many_loci_reads = int(log_dict['Number of reads mapped to too many loci'])
        total_reads = int(log_dict['Number of input reads'])
        multi_reads = multiple_loci_reads + too_many_loci_reads

        self.add_metric(
            name='Genome',
            value=self.genome_name,
        )
        self.add_metric(
            name=f'Uniquely Mapped {self.stat_prefix}',
            value=unique_reads,
            total=total_reads,
            help_info='reads that mapped uniquely to the genome'
        )
        self.add_metric(
            name=f'Multi-Mapped {self.stat_prefix}',
            value=multi_reads,
            total=total_reads,
            help_info='reads that mapped to multiple locations in the genome'
        )


def get_opts_star_mixin(parser, sub_program):
    parser.add_argument(
        '--genomeDir',
        help=HELP_DICT['genomeDir'],
    )
    parser.add_argument(
        '--outFilterMatchNmin',
        help="""Alignment will be output only if the number of matched bases 
is higher than or equal to this value.""",
        default=50,
    )
    parser.add_argument(
        '--out_unmapped',
        help='Output unmapped reads.',
        action='store_true'
    )
    parser.add_argument('--STAR_param', help=HELP_DICT['additional_param'], default="")
    parser.add_argument(
        '--outFilterMultimapNmax',
        help=(
            'How many places are allowed to match a read at most. Please note that even if this value is greater than 1, '
            'at most 1 alignment(with the highest score) will be output in the bam file.'
        ),
        default=1,
    )
    parser.add_argument(
        '--starMem',
        help='Default `30`. Maximum memory that STAR can use.',
        default=30
    )
    if sub_program:
        parser.add_argument('--fq', help="Required. R2 fastq file.", required=True)
        parser.add_argument("--consensus_fq", action='store_true',
                            help="A indicator that the input fastq has been consensused.")
        parser = s_common(parser)



import sys
import subprocess
from collections import Counter

import pandas as pd
import pysam
from scipy import interpolate
import numpy as np
import math

from celescope.tools.__init__ import PATTERN_DICT, FILTERED_MATRIX_DIR_SUFFIX, COUNTS_FILE_NAME 
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.tools.barcode import Chemistry, Barcode
from celescope.tools import utils
from celescope.tools.make_ref import MakeRef
from celescope.tools.matrix import CountMatrix
from celescope.tools.emptydrop_cr import get_plot_elements
from celescope.tools.cells import Cells_metrics

SAM_attributes = 'NH HI nM AS CR UR CB UB GX GN '
MIN_CELL = 500
MAX_CELL = 60000
MIN_BEAD = 100000
MAX_BEAD = 300000

class Starsolo(Step):
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        fq1_list = args.fq1.split(",")
        fq2_list = args.fq2.split(",")
        fq1_number = len(fq1_list)
        fq2_number = len(fq2_list)
        if fq1_number != fq2_number:
            sys.exit('fastq1 and fastq2 do not have same file number!')
        self.read_command = 'cat'
        if str(fq1_list[0]).endswith('.gz'):
            self.read_command = 'zcat'
        if args.chemistry == 'auto':
            ch = Chemistry(args.fq1)
            chemistry_list = ch.check_chemistry()
            if len(set(chemistry_list)) != 1:
                sys.exit('multiple chemistry found!' + str(chemistry_list))
            chemistry = chemistry_list[0]
        else:
            chemistry = args.chemistry
        self.chemistry = chemistry
        
        if chemistry != 'customized':
            self.whitelist_str = " ".join(Chemistry.get_whitelist(chemistry))
            pattern = PATTERN_DICT[chemistry]
        else:
            self.whitelist_str = args.whitelist
            pattern = self.args.pattern
        self.cb_pos, self.umi_pos, self.umi_len = self.get_solo_pos(pattern)
        self.pattern = pattern
       
        # output files
        self.solo_out_dir = f'{self.outdir}/{self.sample}_Solo.out/'
        solo_dir = f'{self.outdir}/{self.sample}_Solo.out/GeneFull_Ex50pAS'
        self.raw_matrix = f'{solo_dir}/raw'
        self.filtered_matrix = f'{solo_dir}/filtered'
        self.summary_file = f'{solo_dir}/Summary.csv'
        bam = f'{self.outdir}/{self.sample}_Aligned.sortedByCoord.out.bam'

        # outs
        self.outs = [self.raw_matrix, self.filtered_matrix, bam]
        
    
    @staticmethod
    def get_solo_pos(pattern):
        # returns: cb_pos, umi_pos, umi_len
        pattern_dict = Barcode.parse_pattern(pattern)
        cb_pos = ' '.join([f'0_{l}_0_{r-1}' for l, r in pattern_dict["C"]])
        if len(pattern_dict['U']) != 1:
            sys.exit(f'Error: Wrong pattern:{pattern}. \n Solution: fix pattern so that UMI only have 1 position.\n')
        ul, ur = pattern_dict["U"][0]
        umi_pos = f'0_{ul}_0_{ur-1}'
        umi_len = ur - ul
        return cb_pos, umi_pos, umi_len

    def run_starsolo(self):
        sa = SAM_attributes + self.args.SAM_attributes
        cmd = (
            'STAR \\\n'
            f'--genomeDir {self.args.genomeDir} \\\n'
            f'--readFilesIn {self.args.fq2} {self.args.fq1} \\\n'
            f'--readFilesCommand {self.read_command} \\\n'
            f'--soloCBwhitelist {self.whitelist_str} \\\n'
            f'--soloType CB_UMI_Complex \\\n'
            f'--soloCBposition {self.cb_pos} \\\n'
            f'--soloUMIposition {self.umi_pos} \\\n'
            f'--soloUMIlen {self.umi_len} \\\n'
            f'--soloCellFilter {self.args.soloCellFilter} \\\n'
            f'--outFileNamePrefix {self.out_prefix}_ \\\n'
            f'--runThreadN {self.thread} \\\n'
            f'--clip3pAdapterSeq {self.args.adapter_3p} \\\n'
            f'--outFilterMatchNmin {self.args.outFilterMatchNmin} \\\n'
            f'--soloFeatures {self.args.soloFeatures} \\\n'
            f'--outSAMattributes {sa} \\\n'
            '--soloCBmatchWLtype 1MM \\\n'
            '--outSAMtype BAM SortedByCoordinate \\\n'
            '--soloCellReadStats Standard \\\n'
        )
        if self.args.STAR_param:
            cmd += self.args.STAR_param
        sys.stderr.write(cmd)
        subprocess.check_call(cmd, shell=True)
        cmd = f'chmod -R 755 {self.solo_out_dir}'
        sys.stderr.write(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def gzip_matrix(self):
        cmd = f'gzip {self.raw_matrix}/*; gzip {self.filtered_matrix}/*'
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def get_Q30_cb_UMI(self):
        fq1_list = self.args.fq1.split(",")
        pattern_dict = Barcode.parse_pattern(self.pattern)
        cb_10k, umi_10k, cb_10k_100k, umi_10k_100k = Counter(), Counter(), Counter(), Counter()
        n = 0
        with pysam.FastxFile(fq1_list[0], persist=False) as fq1:
            for entry in fq1:
                n += 1
                if n > 10 ** 5:
                    break
                qual = entry.quality
                cb_qual = Barcode.get_seq_str(qual, pattern_dict['C'])
                umi_qual = Barcode.get_seq_str(qual, pattern_dict['U'])
                if n <= 10 ** 4:
                    cb_10k.update(cb_qual)
                    umi_10k.update(umi_qual)
                else:
                    cb_10k_100k.update(cb_qual)
                    umi_10k_100k.update(umi_qual)
        
        cb_qual_counter = cb_10k
        umi_qual_counter = umi_10k
        if cb_10k_100k:
            cb_qual_counter = cb_10k_100k
            umi_qual_counter = umi_10k_100k
        q30_cb = sum([cb_qual_counter[k] for k in cb_qual_counter if k >= Barcode.ord2chr(
            30)]) / float(sum(cb_qual_counter.values()))
        q30_umi = sum([umi_qual_counter[k] for k in umi_qual_counter if k >= Barcode.ord2chr(
            30)]) / float(sum(umi_qual_counter.values()))
        return q30_cb, q30_umi


    def run(self):
        self.run_starsolo()
        self.gzip_matrix()
        q30_cb, q30_umi = self.get_Q30_cb_UMI()
        return q30_cb, q30_umi, self.chemistry


def starsolo(args):

    with Starsolo(args) as runner:
        q30_cb, q30_umi, chemistry = runner.run()

    with Mapping(args) as runner:
        valid_reads, corrected = runner.run()
    
    with Cells(args) as runner:
        n_reads, q30_RNA = runner.run(chemistry, valid_reads)
    
    with Demultiplexing(args) as runner:
        runner.run(valid_reads, n_reads, corrected, q30_cb, q30_umi, q30_RNA)


class Mapping(Step):
    # only add metrics
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        solo_dir = f'{self.outdir}/{self.sample}_Solo.out/GeneFull_Ex50pAS'
        self.cellReadsStats = f'{solo_dir}/CellReads.stats'
        self.filtered_matrix = f'{self.outs_dir}/{FILTERED_MATRIX_DIR_SUFFIX}'
        self.counts_file = f'{solo_dir}/{COUNTS_FILE_NAME }'
        self.genome = MakeRef.get_config(args.genomeDir)['meta']['genome_name']

        self.outs = [self.counts_file]


    @utils.add_log
    def run(self):
        df = pd.read_csv(self.cellReadsStats, sep='\t', header=0, index_col=0)
        df = df.iloc[1:,] # skip first line cb not pass whitelist
        df_count = df.loc[:,['nUMIunique','countedU']] # keep dataframe format
        df_count.rename(columns={'nUMIunique': 'UMI'}, inplace=True)
        df = df.loc[:,['cbMatch', 'cbPerfect','genomeU', 'genomeM', 'exonic', 'intronic','exonicAS','intronicAS','countedU']]
        s= df.sum()
        # json does not recognize NumPy data types. TypeError: Object of type int64 is not JSON serializable
        valid = int(s['cbMatch'])
        perfect = int(s['cbPerfect'])
        corrected = valid - perfect
        genomeU= int(s['genomeU'])
        genomeM = int(s['genomeM'])
        exonic = int(s['exonic'])
        intronic = int(s['intronic'])
        antisense = int(s['exonicAS'] + s['intronicAS'])
        intergenic = genomeM + genomeU - exonic - intronic - antisense
        countedU = int(s['countedU'])
        del df

        self.add_metric(
            name='genome',
            value=self.genome,
        )

        self.add_metric(
            name='Reads mapped to unique loci',
            value=genomeU / valid,
            value_type='fraction',
            help_info='Reads that mapped uniquely to the genome.'
        )

        self.add_metric(
            name='Reads mapped to multiple loci',
            value=genomeM / valid,
            value_type='fraction',
            help_info='Reads that mapped to multiple loci in the genome'
        )
        unique_transcriptome = countedU / valid
        self.add_metric(
            name='Reads mapped uniquely to Transcriptome',
            value=unique_transcriptome,
            value_type='fraction',
            help_info='Reads that mapped to a unique gene in the transcriptome. These reads are used for UMI counting.'
        )
        self.add_metric(
            name='Reads assigned to exonic regions',
            value=exonic / valid,
            value_type='fraction',
            help_info='Reads that assigned to exonic regions of genes',
        )
        self.add_metric(
            name='Reads assigned to intronic regions',
            value=intronic / valid,
            value_type='fraction',
            help_info='Reads that assigned to intronic regions of genes',
        )
        self.add_metric(
            name='Reads assigned to intergenic regions',
            value=intergenic / valid,
            value_type='fraction',
            help_info='Reads that can not be assigned to a gene will be considered as intergenic reads.',
        )
        self.add_metric(
            name='Reads assigned Antisense to gene',
            value=antisense / valid,
            value_type='fraction',
            help_info='Reads that assigned to the opposite strand of genes',
        )

        df_count.sort_values(by='UMI', ascending=False, inplace=True)
        cbs = CountMatrix.read_barcodes(self.filtered_matrix)
        df_count['mark'] = 'UB'
        for cb in cbs:
            df_count.loc[cb, 'mark'] = 'CB'
        df_count.to_csv(self.counts_file, sep='\t', index=True)

        return valid, corrected


class Cells(Cells_metrics):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        solo_dir = f'{self.outdir}/{self.sample}_Solo.out/GeneFull_Ex50pAS'
        self.summary_file = f'{solo_dir}/Summary.csv'
        self.counts_file = f'{self.outs_dir}/{COUNTS_FILE_NAME}'


    @utils.add_log
    def parse_summary_add_metrics(self, valid_reads):
        df = pd.read_csv(self.summary_file, index_col=0, header=None)
        s = df.iloc[:,0]
        n_cells = int(s["Estimated Number of Cells"])
        fraction_reads_in_cells = float(s["Fraction of Unique Reads in Cells"])
        mean_used_reads_per_cell = int(s["Mean Reads per Cell"])
        median_umi_per_cell  = int(s["Median UMI per Cell"])
        median_genes_per_cell = int(s["Median GeneFull_Ex50pAS per Cell"])
        total_genes = int(s["Total GeneFull_Ex50pAS Detected"])
        saturation = float(s["Sequencing Saturation"])
        n_reads = int(s["Number of Reads"])
        q30_RNA = float(s["Q30 Bases in RNA read"])

        self.add_cells_metrics(n_cells, fraction_reads_in_cells, mean_used_reads_per_cell, median_umi_per_cell, total_genes, median_genes_per_cell, saturation, valid_reads)
        return n_reads, q30_RNA, fraction_reads_in_cells

    def add_curve_metrics(self, chemistry):
        """
        Returns:
            low barcodes
            loss of cliff
            loss of knee
        """
        df = pd.read_csv(self.counts_file, header=0,sep='\t',index_col=0)
        total = df.shape[0]
        _x = list(range(1, total+1))
        _x = np.log10(_x)
        _y = df['UMI']
        _y = np.log10(_y+1)
        b = interpolate.splrep(_x, _y, s=10)
        d1 = interpolate.splev(_x,b,der=1)
        d2 = interpolate.splev(_x,b,der=2)
        def get_d1_min(x1,x2):
            """
                x1,x2: start and end bc index of the curve
                returns: bc_index, minimum d1
            """
            i = np.argmin(d1[x1:x2+1]) + x1
            return i, d1[i]

        def get_knee_point(x1, x2):
            """
                cliff point(first knee)
                returns: bc_index
            """
            curvature = d2/(1 + d1 ** 2) ** 1.5
            i = np.argmin(curvature[x1:x2+1]) + x1
            return i

        self.add_metric(
            name='Total barcodes',
            value=total,
            help_info='Total deteced barcodes in whitelist.',
            show=False,
        )
        bc_total = math.inf
        if chemistry.startswith('scopeV2'):
            bc_total = 96 ** 3
        elif chemistry.startswith('scopeV3'):
            bc_total = 96 ** 3 * 2
        min_total = bc_total / 2
        low_barcodes = total < min_total
        self.add_metric(
            name='Low total barcodes',
            value=str(low_barcodes),
            show=False,
            help_info='Total number of detected barcodes is lower than expected. This can can be caused by a sample clog.'
        )

        if total < MAX_BEAD:
            sys.stderr.write(f'Warning: Total number of detected barcodes is less than {MAX_BEAD}.\n')
            return True, True, True
        d1_cliff_cell, d1_cliff = get_d1_min(MIN_CELL, MAX_CELL)
        d1_knee_cell, d1_knee = get_d1_min(MIN_BEAD, MAX_BEAD)
        cliff_cell = get_knee_point(MIN_CELL, MAX_CELL)
        self.add_metric(
            name='Minimum derivative of cliff',
            value=round(d1_cliff, 4),
            show=False,
        )
        self.add_metric(
            name='Barcode number at minimum derivative of cliff',
            value=int(d1_cliff_cell),
            show=False,
        )
        self.add_metric(
            name='Barcode number at cliff',
            value=int(cliff_cell),
            show=False,
            help_info='Using the method from emptyDrops to determine the first knee. It is not accurate when the sample shows loss of cliff.'
        )
        loss_of_cliff = d1_cliff > -2
        self.add_metric(
            name='Loss of cliff',
            value=str(loss_of_cliff),
            show=False,
            help_info='If the minimum first derivative around cliff is greater than -2, it is considered as loss of cliff (loss of single cell behavior).'
        )
        self.add_metric(
            name='Minimum derivative of knee',
            value=round(d1_knee, 4),
            show=False,
        )
        self.add_metric(
            name='Barcode number at minimum derivative of knee',
            value=int(d1_knee_cell),
            show=False,
        )
        loss_of_knee = d1_knee > -2
        self.add_metric(
            name='Loss of knee',
            value=str(loss_of_knee),
            show=False,
            help_info='If the minimum first derivative around knee is greater than -2, it is considered as loss of knee.'
        )

        return low_barcodes, loss_of_cliff, loss_of_knee

    @staticmethod
    def get_curve_quality(fraction_reads_in_cells, low_barcodes, loss_of_cliff, loss_of_knee):
        """

        """
        if any([loss_of_cliff, low_barcodes, loss_of_knee]):
            return "fail"
        if fraction_reads_in_cells >= 0.88:
            return 'super'
        elif fraction_reads_in_cells >= 0.75:
            return 'good'
        return 'pass'


    def run(self, chemistry, valid_reads):
        n_reads, q30_RNA, fraction_reads_in_cells = self.parse_summary_add_metrics(valid_reads)
        self.add_data(chart=get_plot_elements.plot_barcode_rank(self.counts_file))
        low_barcodes, loss_of_cliff, loss_of_knee = self.add_curve_metrics(chemistry)
        quality = self.get_curve_quality(fraction_reads_in_cells, low_barcodes, loss_of_cliff, loss_of_knee)
        self.add_metric(
            name='Barcode rank curve quality', 
            value=quality,
            show=False,
        )
        return n_reads, q30_RNA


class Demultiplexing(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

    def run(self, valid_reads, n_reads, corrected, q30_cb, q30_umi, q30_RNA):
        self.add_metric(
            name='Raw Reads',
            value=n_reads,
            help_info='total reads from FASTQ files'
        )
        self.add_metric(
            name='Valid Reads',
            value=valid_reads / n_reads,
            value_type='fraction',
            help_info='fraction of reads with valid barcode and UMI'
        )

        self.add_metric(
            name='Corrected Barcodes',
            value=corrected / valid_reads,
            value_type='fraction',
            help_info='fraction of valid reads with corrected barcodes. Barcodes are corrected to the whitelist sequence that is 1 Hamming-distance away.'
        )

        self.add_metric(
            name='Q30 of Barcodes',
            value=q30_cb,
            value_type='fraction',
            help_info='Fraction of barcode bases with quality score >= 30',
        )

        self.add_metric(
            name='Q30 of UMI',
            value=q30_umi,
            value_type='fraction',
            help_info='Fraction of UMI bases with quality score >= 30',
        )

        self.add_metric(
            name='Q30 of RNA Reads',
            value=q30_RNA,
            value_type='fraction',
            help_info='Fraction of RNA read bases with quality score >= 30',
        )


def get_opts_starsolo(parser, sub_program=True):
    parser.add_argument(
        '--chemistry',
        help='Predefined (pattern, barcode whitelist, linker whitelist) combinations. ' + HELP_DICT['chemistry'],
        choices=list(PATTERN_DICT.keys()),
        default='auto'
    )
    parser.add_argument(
        '--pattern',
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
        - `C`: cell barcode  
        - `L`: linker(common sequences)  
        - `U`: UMI    
        - `T`: poly T""",
    )
    parser.add_argument(
        '--whitelist',
        help='Cell barcode whitelist file path, one cell barcode per line.'
    )
    parser.add_argument(
        '--adapter_3p',
        help='Adapter sequence to clip from 3 prime. Multiple sequences are seperated by space',
        default='AAAAAAAAAAAA',
    )
    parser.add_argument(
        '--genomeDir',
        help=HELP_DICT['genomeDir'],
    )
    parser.add_argument(
        '--outFilterMatchNmin',
        help="""Alignment will be output only if the number of matched bases 
is higher than or equal to this value.""",
        default=50,
    )
    parser.add_argument(
        '--soloCellFilter',
        help='The same as the argument in STARsolo',
        default='EmptyDrops_CR 3000 0.99 10 45000 90000 500 0.01 20000 0.001 10000',
    )
    parser.add_argument(
        '--starMem',
        help='Maximum memory that STAR can use.',
        default=32
    )
    parser.add_argument('--STAR_param', help=HELP_DICT['additional_param'], default="")
    parser.add_argument(
        '--SAM_attributes', 
        help=f'Additional attributes(other than {SAM_attributes}) to be added to SAM file',
        default="")
    parser.add_argument(
        '--soloFeatures', 
        help='The same as the argument in STARsolo',
        default='Gene GeneFull_Ex50pAS',
    )
    if sub_program:
        parser.add_argument('--fq1', help='R1 fastq file. Multiple files are separated by comma.', required=True)
        parser.add_argument('--fq2', help='R2 fastq file. Multiple files are separated by comma.', required=True)
        parser = s_common(parser)

    return parser


import os
import unittest
from collections import namedtuple

from celescope.tools.step import Step


class Tests(unittest.TestCase):

    def setUp(self):
        pass

    @unittest.skip("tested")
    def test_stat_to_metric(self):
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/multi_tests/rna')
        args_dict = {
            'sample': 'test1',
            'assay': 'rna',
            'thread': 1,
            'outdir': 'test1/06.analysis',
            'debug': True,
        }
        Args = namedtuple('Args', list(args_dict.keys()))
        args = Args(**args_dict)

        obj = Step(args, 'analysis')
        obj.stat_to_metric()
        print(obj.__content_dict['metric'])

    def test_test(self):
        assert 0 == 0




import os
import pathlib
from collections import defaultdict
from itertools import groupby
import subprocess
import unittest
import shutil

import pysam

from celescope.rna.mkref import Mkref_rna
from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.tools import reference
from celescope.tools.__init__ import TAG_BAM_SUFFIX


GTF_TYPES = ['exon','gene']


def correct_umi(umi_dict, percent=0.1):
    """
    Correct umi_dict in place.
    Args:
        umi_dict: {umi_seq: umi_count}
        percent: if hamming_distance(low_seq, high_seq) == 1 and
            low_count / high_count < percent, merge low to high.
        return_dict: if set, return correct_dict = {low_seq:high_seq}
    Returns:
        n_corrected_umi: int
        n_corrected_read: int
    """
    n_corrected_umi = 0
    n_corrected_read = 0
    dic = {}

    # sort by value(UMI count) first, then key(UMI sequence)
    umi_arr = sorted(
        umi_dict.items(), key=lambda kv: (kv[1], kv[0]), reverse=True)
    while True:
        # break when only highest in umi_arr
        if len(umi_arr) == 1:
            break
        umi_low = umi_arr.pop()
        low_seq = umi_low[0]
        low_count = umi_low[1]

        for umi_kv in umi_arr:
            high_seq = umi_kv[0]
            high_count = umi_kv[1]
            if float(low_count / high_count) > percent:
                break
            if utils.hamming_distance(low_seq, high_seq) == 1:
                n_low = umi_dict[low_seq]
                n_corrected_umi += 1
                n_corrected_read += n_low
                # merge
                umi_dict[high_seq] += n_low
                dic[low_seq] = high_seq
                del (umi_dict[low_seq])
                break

    return n_corrected_umi, n_corrected_read, dic


def discard_read(gene_umi_dict):
    """
    If two or more groups of reads have the same barcode and UMI, but different gene annotations, the gene annotation with the most supporting reads is kept for UMI counting, and the other read groups are discarded. In case of a tie for maximal read support, all read groups are discarded, as the gene cannot be confidently assigned.

    Returns:
        discarded_umi: set. umi with tie read count
        umi_gene_dict: {umi_seq: {gene_id: read_count}}
    """

    discard_umi = set()
    umi_gene_dict = defaultdict(lambda: defaultdict(int))
    for gene_id in gene_umi_dict:
        for umi in gene_umi_dict[gene_id]:
            umi_gene_dict[umi][gene_id] += gene_umi_dict[gene_id][umi]
    
    for umi in umi_gene_dict:
        max_read_count = max(umi_gene_dict[umi].values())
        gene_id_max = [gene_id for gene_id, read_count in umi_gene_dict[umi].items() if read_count==max_read_count]

        if len(gene_id_max) > 1:
            discard_umi.add(umi)
        else:
            gene_id = gene_id_max[0]
            umi_gene_dict[umi] = {gene_id: umi_gene_dict[umi][gene_id]}

    return discard_umi, umi_gene_dict


class FeatureCounts(Step):

    """
    ## Features
    - Assigning uniquely mapped reads to genomic features with FeatureCounts.
    ## Output
    - `{sample}` Numbers of reads assigned to features (or meta-features).
    - `{sample}_summary` Stat info for the overall summrization results, including number of 
    successfully assigned reads and number of reads that failed to be assigned due to 
    various reasons (these reasons are included in the stat info).
    - `{sample}_aligned_sortedByCoord_addTag.bam` featureCounts output BAM, 
    sorted by coordinates
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        # set
        self.gtf = Mkref_rna.get_config(self.args.genomeDir)['files']['gtf']
        gp = reference.GtfParser(self.gtf)
        self.id_name = gp.get_id_name()
        self.intron_dict = {}

        # stats
        self.exon = self.intron = self.intergenic = self.ambiguity = 0


        # temp file
        self.tmp_dir = f'{self.outdir}/tmp/'
        self.add_tag_bam = f'{self.out_prefix}_addTag.bam'
        input_basename = os.path.basename(self.args.input)
        self.exon_bam = f'{self.tmp_dir}/exon/{input_basename}.featureCounts.bam'
        self.intron_bam = f'{self.tmp_dir}/intron/{input_basename}.featureCounts.bam'

        # out
        self.count_detail_file = f'{self.out_prefix}_count_detail.txt'
        self.nameSorted_bam = f'{self.out_prefix}_nameSorted.bam'
        self.out_bam = f'{self.out_prefix}_{TAG_BAM_SUFFIX}'

    def add_tag(self, seg, id_name):
        """
        Add intron reads and tag

        Args:
            seg: pysam bam segment
            id_name: {gene_id: gene_name}

        Returns:
            seg with tag added

        Tags:
            CB: cell barcode
            UB: error-corrected UMI
            UR: original UMI
            GN: gene name
            GX: gene_id

        """
        attr = seg.query_name.split(':')
        barcode = attr[0]
        ur = ub = attr[1]

        # assign to some gene
        xs = seg.get_tag('XS')
        if xs == 'Assigned':
            gene_id = seg.get_tag('XT')
            gene_name = id_name[gene_id]
            seg.set_tag(tag='GN', value=gene_name, value_type='Z')
            seg.set_tag(tag='GX', value=gene_id, value_type='Z')
            seg.set_tag(tag='RE', value='E', value_type='Z')
            self.exon += 1
        else:
            if self.intron_dict and seg.query_name in self.intron_dict:
                gene_id = self.intron_dict[seg.query_name]
                gene_name = id_name[gene_id]
                seg.set_tag(tag='GN', value=gene_name, value_type='Z')
                seg.set_tag(tag='GX', value=gene_id, value_type='Z')
                seg.set_tag(tag='RE', value='N', value_type='Z')
                seg.set_tag(tag='XT', value=gene_id, value_type='Z')
                self.intron += 1
            elif xs == 'Unassigned_NoFeatures':
                seg.set_tag(tag='RE', value='I', value_type='Z')
                self.intergenic += 1
            elif xs == 'Unassigned_Ambiguity':
                seg.set_tag(tag='RE', value='A', value_type='Z')
                self.ambiguity += 1
        
        seg.set_tag(tag='CB', value=barcode, value_type='Z')
        seg.set_tag(tag='UB', value=ub, value_type='Z')
        seg.set_tag(tag='UR', value=ur, value_type='Z')

        return seg

    @utils.add_log
    def get_intron_dict(self):
        with pysam.AlignmentFile(self.intron_bam, "rb") as in_bam:
            for seg in in_bam:
                if seg.has_tag('XT'):
                    self.intron_dict[seg.query_name] = seg.get_tag('XT')


    @utils.add_log
    def run_featureCounts(self, outdir, gtf_type):
        '''
        allow multimapping with -M; but each multi-mapped reads only have one alignment because of --outSAMmultNmax 1
        '''
        cmd = (
            'featureCounts '
            f'-s 1 '
            f'--largestOverlap '
            f'-M '
            f'-a {self.gtf} '
            f'-o {outdir}/{self.sample} '  
            '-R BAM '
            f'-T {self.args.thread} '
            f'-t {gtf_type} '
            f'{self.args.input} '
            '2>&1 '
        )
        if self.args.featureCounts_param:
            cmd += (" " + self.args.featureCounts_param)
        self.run_featureCounts.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    def run_exon_intron(self):
        tmp_dir = f'{self.outdir}/tmp/'
        for gtf_type in ['exon', 'intron']:
            outdir = f'{tmp_dir}/{gtf_type}'
            pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)            
            self.run_featureCounts(outdir, gtf_type)

    def remove_temp_file(self):
        shutil.rmtree(self.tmp_dir)
        os.remove(self.add_tag_bam)

    def run(self):
        self.run_exon_intron()
        self.get_intron_dict()
        utils.sort_bam(
            self.exon_bam,
            self.nameSorted_bam,
            threads=self.thread, by='name')
        self.get_count_detail_add_tag()
        self.add_metrics()
        utils.sort_bam(
            input_bam=self.add_tag_bam,
            output_bam=self.out_bam)
        self.remove_temp_file()

    @utils.add_log
    def add_metrics(self):
        total = self.exon + self.intron + self.intergenic + self.ambiguity

        self.add_metric(
            name='Feature Type',
            value=self.args.gtf_type.capitalize(),
            help_info='Specified by `--gtf_type`. For snRNA-seq, you need to add `--gtf_type gene` to include reads mapped to intronic regions. Staring from CeleScope v1.12.0, the default value of gtf_type is changed from `exon` to `gene`.'
        )
        self.add_metric(
            name='Reads Assigned To Exonic Regions',
            value=self.exon,
            total=total,
            help_info='Reads that can be successfully assigned to exonic regions'
        )
        self.add_metric(
            name='Reads Assigned To Intronic Regions',
            value=self.intron,
            total=total,
            help_info='Reads that can be successfully assigned to intronic regions'
        )
        self.add_metric(
            name='Reads Assigned To Intergenic Regions',
            value=self.intergenic,
            total=total,
            help_info='Reads that can be successfully assigned to intergenic regions'
        )
        self.add_metric(
            name='Reads Unassigned Ambiguity',
            value=self.ambiguity,
            total=total,
            help_info='Alignments that overlap two or more features'
        )

    @utils.add_log
    def get_count_detail_add_tag(self):
        """
        bam to detail table
        must be used on name_sorted bam
        Output file:
            - count_detail_file
            - bam with tag(remain name sorted)
        """
        save = pysam.set_verbosity(0)
        inputFile = pysam.AlignmentFile(self.nameSorted_bam, "rb")
        outputFile = pysam.AlignmentFile(self.add_tag_bam, 'wb', header=inputFile.header)
        pysam.set_verbosity(save)

        with open(self.count_detail_file, 'wt') as fh1:
            fh1.write('\t'.join(['Barcode', 'geneID', 'UMI', 'read', 'unique', 'PCR_duplicate']) + '\n')

            def keyfunc(x):
                return x.query_name.split(':', 1)[0]
            for _, g in groupby(inputFile, keyfunc):
                gene_umi_dict = defaultdict(lambda: defaultdict(int))
                gene_umi_pos = utils.genDict(dim=3, valType=int)
                for seg in g:
                    seg = self.add_tag(seg, self.id_name)
                    outputFile.write(seg)
                    barcode, umi = seg.get_tag('CB'), seg.get_tag('UB')
                    if not seg.has_tag('GX'):
                        continue
                    gene_id = seg.get_tag('GX')
                    gene_umi_dict[gene_id][umi] += 1
                    gene_umi_pos[gene_id][umi][seg.reference_start] += 1                

                # output
                for gene_id in gene_umi_dict:
                    n_umi = len(gene_umi_dict[gene_id])
                    n_read = 0
                    unique = dup = 0
                    for umi in gene_umi_dict[gene_id]:
                        read_count = gene_umi_dict[gene_id][umi] 
                        n_read += read_count
                        if read_count == 1:
                            # unique
                            unique += 1
                        else:
                            # only add postion duplicate read number
                            for pos in gene_umi_pos[gene_id][umi]:
                                if gene_umi_pos[gene_id][umi][pos] > 1:
                                    dup += gene_umi_pos[gene_id][umi][pos]
                    fh1.write(f'{barcode}\t{gene_id}\t{n_umi}\t{n_read}\t{unique}\t{dup}\n')

        inputFile.close()
        outputFile.close()


@utils.add_log
def featureCounts(args):
    with FeatureCounts(args) as runner:
        runner.run()


def get_opts_featureCounts(parser, sub_program):
    parser.add_argument(
        '--gtf_type',
        help='Specify feature type in GTF annotation',
        default='gene',
        choices=['exon', 'gene'],
    )
    parser.add_argument('--genomeDir', help=HELP_DICT['genomeDir'])
    parser.add_argument('--featureCounts_param', help=HELP_DICT['additional_param'], default="")
    #parser.add_argument('--correct_UMI', help='perform UMI correction.')

    if sub_program:
        parser.add_argument('--input', help='Required. BAM file path.', required=True)
        parser = s_common(parser)
    return parser


class featureCounts_test(unittest.TestCase):
    def test_correct_umi(self):
        dic = {
            "apple1": 2,
            "apple2": 30,
            "bears1": 5,
            "bears2": 10,
            "bears3": 100,
            "ccccc1": 20,
            "ccccc2": 199,
        }
        n_corrected_umi, n_corrected_read, _ = correct_umi(dic)
        dic_after_correct = {
            'ccccc1': 20,
            'apple2': 32,
            'bears3': 115,
            'ccccc2': 199,
        }
        self.assertEqual(dic, dic_after_correct)
        self.assertEqual(n_corrected_umi, 3)
        self.assertEqual(n_corrected_read, 2 + 5 + 10)

if __name__ == "__main__":
    unittest.main()


from celescope.tools import utils
from celescope.tools.star_mixin import Star_mixin, get_opts_star_mixin

class Star(Star_mixin):
    """
    ## Features
    - Align R2 reads to the reference genome with STAR.

    ## Output
    - `{sample}_Aligned.sortedByCoord.out.bam` BAM file contains Uniquely Mapped Reads.

    - `{sample}_SJ.out.tab` SJ.out.tab contains high confidence collapsed splice junctions in tab-delimited format.

    - `{sample}_Log.out` Main log with a lot of detailed information about the run. 
    This is most useful for troubleshooting and debugging.

    - `{sample}_Log.progress.out` Report job progress statistics, such as the number of processed reads, 
    % of mapped reads etc. It is updated in 1 minute intervals.

    - `{sample}_Log.Log.final.out` Summary mapping statistics after mapping job is complete, 
    very useful for quality control. The statistics are calculated for each read (single- or paired-end) and 
    then summed or averaged over all reads. Note that STAR counts a paired-end read as one read, 
    (unlike the samtools agstat/idxstats, which count each mate separately). 
    Most of the information is collected about the UNIQUE mappers 
    (unlike samtools agstat/idxstats which does not separate unique or multi-mappers). 
    Each splicing is counted in the numbers of splices, which would correspond to 
    summing the counts in SJ.out.tab. The mismatch/indel error rates are calculated on a per base basis, 
    i.e. as total number of mismatches/indels in all unique mappers divided by the total number of mapped bases.
    """

    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)


    
    @utils.add_log
    def run(self):
        super().run()



def star(args):
    with Star(args, display_title="Mapping") as runner:
        runner.run()

def get_opts_star(parser, sub_program):
    get_opts_star_mixin(parser, sub_program)



from celescope.tools.capture.analysis import Analysis, get_opts_analysis


def analysis_virus(args):

    with Analysis_virus(args) as runner:
        runner.run()


def get_opts_analysis_virus(parser, sub_program):
    get_opts_analysis(parser, sub_program)


class Analysis_virus(Analysis):
    pass




# Li Song: the program to re-cluster the CDR3 regions

#!/usr/bin/env python3

import sys
import math

def GetChainType(v, j, c):
	s = ""
	if (v != "*"):
		s = v
	elif (c != "*"):
		s = c 
	elif (j != "*"):
		s = j
	else:
		return -1
	
	if (s[0:3] == "IGH"):
		return 0
	elif (s[0:3] == "IGK"):
		return 1
	elif (s[0:3] == "IGL"):
		return 2
	elif (s[0:3] == "TRA"):
		return 3
	elif (s[0:3] == "TRB"):
		return 4
	elif (s[0:3] == "TRG"):
		return 5
	elif (s[0:3] == "TRD"):
		return 6
	else:
		return -1

def GetMainGeneName(g):
	return g.split("*")[0] 

def CompatibleGeneAssignment(a, b):
	if ( GetChainType( a[0], a[2], a[3] ) != GetChainType(b[0], b[2], b[3]) ):
		return False 
	
	for i in [0, 2]: # V, J gene assignment
		ga = a[i].split("*")[0]
		gb = b[i].split("*")[0]
		if ( ga != "" and gb != "" and ga != gb ):
			return False
	return True

def GetSimilarity(a, b):
	if (len(a) != len(b)):
		return 0
	sameCnt = 0
	for i in range(len(a)):
		if (a[i] == b[i]):
			sameCnt += 1
	return sameCnt / len(a)

def CompatibleSequence(a, b, similarity):
	if (len(a) != len(b)):
		return False ;
	diffCnt = 0 
	diffMax = len(a) - int(math.ceil(len(a) * similarity))
	for i in range(len(a)):
		if (a[i] != b[i]):
			diffCnt += 1
			if (diffCnt > diffMax):
				return False
	return True

def GetFather(tag, father):
	if ( father[tag] != tag ):
		father[tag] = GetFather( father[tag], father )
	return father[tag]
	
def LargerCluster(rawCdr3List, similarity, prefix, useRepresentative, mode):
	vjCDR3LenList = {}
	clusterNameToId = {}
	clusterIdToName = []
	clusterRepresentativeId = {}
	clusterRepresentativeAbund = {}
	
	if (len(rawCdr3List) == 0):
		return
	cdr3List = sorted(rawCdr3List, key=lambda x:(x[0], x[8]))
	# Select the represenative, allowing same CDR3 shown up in different rows in a cluster to handle scRNA-seq
	abund = cdr3List[0][10]
	for i in range(1, len(cdr3List) + 1):
		prevKey = (cdr3List[i - 1][0], cdr3List[i - 1][8])
		key = "*"
		if (i < len(cdr3List)):
			key = (cdr3List[i][0], cdr3List[i][8])
		if (key == prevKey):
			abund += cdr3List[i][10]
		else:
			cdr3 = cdr3List[i - 1]
			if ( cdr3[0] not in clusterNameToId ):
				clusterNameToId[cdr3[0]] = len(clusterIdToName)
				clusterIdToName.append(cdr3[0])
				clusterRepresentativeId[cdr3[0]] = i - 1
				clusterRepresentativeAbund[cdr3[0]] = abund
			elif (abund > clusterRepresentativeAbund[cdr3[0]] ):
				clusterRepresentativeId[cdr3[0]] = i - 1
				clusterRepresentativeAbund[cdr3[0]] = abund
			if (i < len(cdr3List)):
				abund = cdr3List[i][10]
	#print(cdr3List[clusterRepresentativeId["assemble2"]])
	# Reorganize the CDR3 list by their V, J and CDR3 length.
	i = 0
	for cdr3 in cdr3List:
		if (useRepresentative and clusterRepresentativeId[cdr3[0]] != i):
			i += 1
			continue
		key = (GetMainGeneName(cdr3[2]), GetMainGeneName(cdr3[4]), len(cdr3[8]))
		if (key not in vjCDR3LenList):
			vjCDR3LenList[key] = []
		vjCDR3LenList[key].append(i)
		i += 1
	# Prepare the set-union
	father = []
	for cdr3 in cdr3List:
		father.append(clusterRepresentativeId[cdr3[0]])
	
	# Build up the set-union relation
	if (mode == "aggressive"):
		for key in vjCDR3LenList.keys():
			cdr3IdList = vjCDR3LenList[key]
			size = len( cdr3IdList )
			for i in range(size):
				fi = GetFather(cdr3IdList[i], father)
				for j in range(i + 1, size):
					fj = GetFather(cdr3IdList[j], father) 
					if ( fi != fj  
						and CompatibleSequence( cdr3List[ cdr3IdList[i] ][8], cdr3List[ cdr3IdList[j] ][8], 
						similarity ) ):
						#if (cdr3List[cdr3IdList[j]][8] == "TGCCAACAGTATATTAGTTACTCGTACACTTTT"):
						#	print(i, cdr3List[cdr3IdList[i]])
						#	print(j, cdr3List[cdr3IdList[j]])
						father[fj] = fi
	elif (mode == "center"):
		for key in vjCDR3LenList.keys():
			rawCdr3IdList = vjCDR3LenList[key][:]
			cdr3IdList = sorted(rawCdr3IdList, key=
					lambda x: (clusterRepresentativeAbund[cdr3List[x][0]], cdr3List[x][10]), reverse=True)
			size = len(cdr3IdList)
			for i in range(1, size):
				maxFj = 0
				maxSimilarity = -1
				fi = GetFather(cdr3IdList[i], father)
				for j in range(i):
					fj = GetFather(cdr3IdList[j], father)
					if (fi == fj):
						continue
					s = GetSimilarity(cdr3List[fi], cdr3List[fj])
					if (s > maxSimilarity):
						maxSimilarity = s
						maxFj = fj
				if (maxSimilarity >= similarity):
					father[fi] = maxFj

	# Associate original clusters to larger cluster
	largerClusterToId = []
	largerClusterToClusterName = []
	rootToLargerClusterId = {}
	i = 0 
	for cdr3 in cdr3List:
		root = GetFather(i, father)
		lcId = 0
		if (root not in rootToLargerClusterId):
			rootToLargerClusterId[root] = len(largerClusterToId)
			lcId = len(largerClusterToId)
			largerClusterToId.append([])
			largerClusterToClusterName.append(set({}))
		else:
			lcId = rootToLargerClusterId[root]

		largerClusterToId[lcId].append(i)
		largerClusterToClusterName[lcId].add(cdr3[0])
		i += 1		
	
	# Output the result
	# Output the composition of larger cluster
	#for i in range(len(largerClusterToId)):
	#	print("#\tcluster" + str(i), end = "")
	#	for name in largerClusterToClusterName[i]:
	#		print("\t" + name, end = "")
	#	print("\n", end = "")
	
	# Output the new cdr3 format with new larger cluster Id.
	for i in range(len(largerClusterToId)):
		j = 0
		for cdr3Id in largerClusterToId[i]:
			cdr3List[cdr3Id].append(cdr3List[cdr3Id][0])
			cdr3List[cdr3Id].append(cdr3List[cdr3Id][1])
			cdr3List[cdr3Id][0] = prefix + "_" + str(i) #+ "_" + cdr3List[cdr3Id][0] + "_" + str(cdr3List[cdr3Id][1])
			cdr3List[cdr3Id][1] = j
			print( "\t".join( str(x) for x in cdr3List[cdr3Id] ) )
			j += 1

	return 


if (__name__ == "__main__"):
	if (len(sys.argv) <= 1):
		print("usage: a.py trust_cdr3.out [OPTIONS] > output\n" + 
			"OPTIONS:\n" +
			"\t-s FLOAT: similarity of two CDR3s (default: 0.8)\n" +
			"\t--prefix STRING: prefix to new cluster name (default: cluster)\n" + 
			"\t--center: use the center of the cluster for similarity comparison (default: no)\n"+
			"\t--representative: use representative CDR3 from each contig for cluster (default: no)\n" +
			"\t--format [cdr3, simplerep]: the input format type (default: cdr3)")
		exit(1)

	cdr3List = []
	similarity = 0.8 ;
	prefix = "cluster"
	useRepresentative = False
	mode = "aggressive"
	inputFormat = "cdr3"
	i = 2
	while (i < len(sys.argv)):
		if (sys.argv[i] == "-s"):
			similarity = float(sys.argv[i + 1])
			i += 1
		elif (sys.argv[i] == "--prefix"):
			prefix = sys.argv[i + 1]
			i += 1
		elif (sys.argv[i] == "--representative"):
			useRepresentative = True
		elif (sys.argv[i] == "--center"):
			mode = "center"
		elif (sys.argv[i] == "--format"):
			inputFormat = sys.argv[i + 1]
			if (inputFormat not in ["cdr3", "simplerep"]):
				print("Unknown format: ", inputFormat)
				exit(1)
			i += 1
		else:
			print("Unknown option: ", sys.argv[i])
			exit(1)
		i += 1

	fp = open(sys.argv[1])
	lineCnt = 0
	for line in fp:
		line = line.rstrip()
		cols = line.split("\t")
		if (inputFormat == "cdr3"):
			cols[1] = int(cols[1])
			skip = False
			for g in [2, 4]: # Must have V, J genes.
				if ( cols[g] == "*"):
					skip = True 
			if (float(cols[9]) == 0):
				skip = True
			if ( skip ):
				continue 
			for	g in [2, 3, 4, 5]:
				cols[g] = cols[g].split(",")[0]
			cols[9] = float(cols[9])
			cols[10] = float(cols[10])
			if (cols[9] == 0):
				continue
		elif (inputFormat == "simplerep"):
			if (line[0] == "#"):
				continue
			if ("_" in cols[3] or "?" in cols[3]):
				continue
			for g in [4, 6]: # Must have V, J genes.
				if ( cols[g] == "*"):
					skip = True 
			reformat = [0] * 11
			reformat[0] = "line" + str(lineCnt)
			reformat[1] = 0
			for	g in [4, 5, 6, 7]:
				reformat[g - 2] = cols[g]	
			reformat[6] = reformat[7] = "*"
			reformat[8] = cols[2]
			reformat[9] = 1
			reformat[10] = cols[0]
			cols = reformat[:]
		cdr3List.append(cols)
		lineCnt += 1
	fp.close()
	LargerCluster(cdr3List, similarity, prefix, useRepresentative, mode)



import subprocess
import os
import warnings

import scanpy as sc
import pandas as pd
import pysam
import matplotlib
import matplotlib.pyplot as plt

from celescope.tools import utils
from celescope.tools.step import Step
from celescope.tools.step import s_common
from celescope.tools.target_metrics import get_gene_list
from celescope.__init__ import HELP_DICT
from celescope.snp.__init__ import PANEL


matplotlib.use('Agg')
warnings.filterwarnings("ignore")

AA_DICT = {
    'Gly' : 'G',
    'Ala' : 'A',
    'Val' : 'V',
    'Leu' : 'L',
    'Ile' : 'I',
    'Phe' : 'F',
    'Trp' : 'W',
    'Tyr' : 'Y',
    'Asp' : 'D',
    'Asn' : 'N',
    'Glu' : 'E',
    'Lys' : 'K',
    'Gln' : 'Q',
    'Met' : 'M',
    'Ser' : 'S',
    'Thr' : 'T',
    'Cys' : 'C',
    'Pro' : 'P',
    'His' : 'H',
    'Arg' : 'R',
}


def parse_variant_ann(variant_ann_file):
    """
    Args:
        variant_ann_file: variant annotation file from snpEff.
    
    Returns:
        gene_list, mRNA_list, protein_list
    """
    gene_list, mRNA_list, protein_list = [], [], []

    with open(variant_ann_file) as f:
        for line in f.readlines():
            if not line.startswith("#"):
                info = line.split('\t')[7]
                anns = info.split("|")
                gene = anns[3]
                gene_list.append(gene)
            
                tmp1, tmp2 = [], []
                for ann in anns:
                    if ann.startswith("c."):
                        exon_loc = anns[anns.index(ann) - 1].split('/')[0]
                        # WARNING_TRANSCRIPT_INCOMPLETE
                        if not exon_loc:
                            continue
                        
                        exon = ann.strip("c.")
                        exon = f"exon{exon_loc}:{exon}"
                        if exon not in tmp1:
                            tmp1.append(exon)

                    if ann.startswith("p."):
                        protein = ann[2:]
                        for i in AA_DICT:
                            protein = protein.replace(i, AA_DICT[i])
                        if protein not in tmp2:
                            tmp2.append(protein)
                        
                mRNA_list.append(','.join(tmp1))
                protein_list.append(','.join(tmp2))

    return (gene_list, mRNA_list, protein_list)


def parse_vcf_to_df(vcf_file, cols=('chrom', 'pos', 'alleles'), infos=('VID', 'CID')):
    """
    Read cols and infos into pandas df
    """
    vcf = pysam.VariantFile(vcf_file)
    df = pd.DataFrame(columns=[col.capitalize() for col in cols] + infos)
    rec_dict = {}
    for rec in vcf.fetch():

        for col in cols:
            rec_dict[col.capitalize()] = getattr(rec, col)
            if col == 'alleles':
                rec_dict['Alleles'] = '-'.join(rec_dict['Alleles'])

        for info in infos:
            rec_dict[info] = rec.info[info]

        '''
        rec_dict['GT'] = [s['GT'] for s in rec.samples.values()][0]
        rec_dict['GT'] = [str(item) for item in rec_dict['GT']]
        rec_dict['GT'] = '/'.join(rec_dict['GT'])
        '''
        df_new = pd.DataFrame(rec_dict, index=[0])
        df = pd.concat([df, df_new])

    vcf.close()
    df.reset_index(drop=True, inplace=True)
    return df


def vcf_to_gt_csv(vcf_file, csv_file):
    vcf = pysam.VariantFile(vcf_file)
    
    samples = vcf.header.samples
    
    with open(csv_file, 'w') as f:
        header = ['variant'] + list(samples)
        f.write(','.join(header) + '\n')
        
        for record in vcf:
            mutation_name = f"{record.chrom}_{record.pos}"
            genotypes = []
            
            for sample in samples:
                genotype = record.samples[sample]['GT']
                g1, g2 = genotype
                
                if g1 is None:
                    genotype_str = "NA"
                else:
                    genotype_str = '/'.join([str(g1),str(g2)])
                
                genotypes.append(genotype_str)
            
            line = [mutation_name] + genotypes
            f.write(','.join(line) + '\n')


class Analysis_snp(Step):
    """
    ## Features
    - Annotate variants with [snpEff](http://pcingola.github.io/SnpEff/).

    ## Output
    - `{sample}_gt.csv` Genotypes of variants of each cell. Rows are variants and columns are cells.
    - `{sample}_variant_ncell.csv` Number of cells with each genotype.
    - `{sample}_variant_table.csv` annotated with snpEff.

    """

    def __init__(self, args, display_title=None):
        super().__init__(args, display_title)
        self.vcf_file = args.vcf

        # parse
        self.gene_list, self.n_gene = get_gene_list(args)
        self.database = args.database

        # data
        self.variant_table = None

        # out
        self.snpeff_outdir = f'{self.outdir}/snpEff/'
        self.snpeff_ann_vcf_file = f'{self.snpeff_outdir}/variants_ann.vcf'
        self.final_vcf_file = f'{self.out_prefix}_final.vcf'
        utils.check_mkdir(self.snpeff_outdir)
        self.plot_snp_dir = f'{self.outdir}/{self.sample}_plot_snp/'

        self.gt_file = f'{self.out_prefix}_gt.csv'
        self.ncell_file = f'{self.out_prefix}_variant_ncell.csv'
        self.variant_table_file = f'{self.out_prefix}_variant_table.csv'

    @utils.add_log
    def write_gt(self):
        vcf_to_gt_csv(self.final_vcf_file, self.gt_file)

    @utils.add_log
    def write_ncell(self):
        """
        parse gt_file to collect each genotype cell count into ncell_file
        """
        df = pd.read_csv(self.gt_file, index_col=0)
        df_ncell = df.apply(pd.Series.value_counts, axis=1).fillna(0).astype(int)
        df_ncell.to_csv(self.ncell_file, index=True)

    @utils.add_log
    def run_snpEff(self):
        cmd = (
            f"snpEff -Xmx8g -v {self.database} {os.path.abspath(self.vcf_file)} > variants_ann.vcf "
        )
        self.run_snpEff.logger.info(cmd)

        cwd = os.getcwd()
        os.chdir(self.snpeff_outdir)
        subprocess.check_call(cmd, shell=True)
        # change dir back to avoid can not find '09.analysis_snp/stat.txt' error
        os.chdir(cwd)

    @utils.add_log
    def keep_in_gene(self):
        """
        Output:
            self.final_vcf_file
        """
        gene_list, _, _ = parse_variant_ann(self.snpeff_ann_vcf_file)
        with pysam.VariantFile(self.snpeff_ann_vcf_file) as vcf_in:
            with pysam.VariantFile(self.final_vcf_file, 'w', header=vcf_in.header) as vcf_out:
                for i, record in enumerate(vcf_in.fetch()):
                    if gene_list[i] in self.gene_list:
                        vcf_out.write(record)              


    def get_variant_table(self):
        """
        Returns:
            is_in_gene_list: if res[i] == True, line i is in gene_list
        """

        df_vcf = parse_vcf_to_df(self.final_vcf_file, infos=[])
        df_vcf["Gene"], df_vcf["mRNA"], df_vcf["Protein"] =  parse_variant_ann(self.final_vcf_file)
        df_ncell = pd.read_csv(self.ncell_file)
        df_vcf = pd.concat([df_vcf, df_ncell], axis=1)

        cols = ["Chrom", "Pos", "Alleles", "Gene", "0/0", "0/1", "1/1", "mRNA", "Protein"]
        cols = [col for col in cols if col in df_vcf.columns]
        df_vcf = df_vcf.loc[:, cols]
        is_in_gene_list = df_vcf.Gene.isin(self.gene_list)
        df_vcf = df_vcf[is_in_gene_list]

        self.variant_table = df_vcf
        self.variant_table.reset_index(drop=True, inplace=True)
        self.variant_table.to_csv(self.variant_table_file, index=False)

    def add_help(self):
        '''
            <p> Chrom : chromosome name.</p>
            <p> Pos : the 1-based position of the variation on the given sequence..</p>
            <p> Alleles : REF(reference base or bases in the case of an indel) - ALT(alternative alleles).</p>
            <p> 0/0, 0/1, 1/1: number of cells with each genotype.</p>
            <p> Gene : gene symbol.</p>
            <p> mRNA :  A standard nomenclature is used in specifying the sequence changes.</p>
            <p> Protein :  A standard nomenclature is used in specifying the sequence changes.</p>
        '''
        self.add_help_content(
            name='Chrom',
            content='Chromosome name'
        )
        self.add_help_content(
            name='Pos',
            content='the 1-based position of the variation on the given sequence'
        )
        self.add_help_content(
            name='Alleles',
            content='REF(reference base or bases in the case of an indel) - ALT(alternative alleles)'
        )
        self.add_help_content(
            name='0/0, 0/1, 1/1',
            content='number of cells with each genotype'
        )
        self.add_help_content(
            name='Gene',
            content='gene symbol'
        )
        self.add_help_content(
            name='mRNA',
            content='A standard nomenclature is used in specifying the sequence changes'
        )
        self.add_help_content(
            name='Protein',
            content='A standard nomenclature is used in specifying the sequence changes'
        )

    @utils.add_log
    def plot_snp(self):
        match_dict = utils.parse_match_dir(self.args.match_dir)
        if 'h5ad' not in match_dict:
            return

        utils.check_mkdir(self.plot_snp_dir)
        df_gt = pd.read_csv(self.gt_file, keep_default_na=False, index_col=0)
        df_v = self.variant_table.copy()
        df_v['n_variants'] = df_v['0/1'] + df_v['1/1']
        indices = df_v.nlargest(self.args.plot_top_n, 'n_variants').index
        df_top = df_gt.iloc[indices,]
        df_top = df_top.transpose()
        variants = df_top.columns
        for c in variants:
            df_top[c] = df_top[c].astype('category')

        adata = sc.read_h5ad(match_dict['h5ad'])
        adata.obs = pd.concat([adata.obs, df_top], axis=1)
        pt_size = min(100, 120000 / len(adata.obs))
        gene_list, protein_list = df_v['Gene'], df_v['Protein']
        for i, v in enumerate(variants):
            title = f'top{i+1}_{variants[i]}_{gene_list[indices[i]]}_{protein_list[indices[i]]}'
            file_name = f'{self.plot_snp_dir}/{title}.pdf'
            sc.pl.umap(adata, color=v, size=pt_size, 
            palette={'0/0':'dimgray', '0/1':'orange', '1/1':'red','NA':'lightgray'},
            title=title)
            plt.savefig(file_name,dpi=300,bbox_inches="tight")



    def run(self):
        self.run_snpEff()
        self.keep_in_gene()
        self.write_gt()
        self.write_ncell()
        self.get_variant_table()
        self.add_help()
        self.plot_snp()
        table_dict = self.get_table_dict(title='Variant table', table_id='variant', df_table=self.variant_table)
        self.add_data(table_dict=table_dict)


@utils.add_log
def analysis_snp(args):
    with Analysis_snp(args, display_title='Analysis') as runner:
        runner.run()


def get_opts_analysis_snp(parser, sub_program):
    parser.add_argument("--gene_list", help=HELP_DICT['gene_list'])
    parser.add_argument("--database", help='snpEff database. Common choices are GRCh38.mane.1.0.ensembl(human) and GRCm38.99(mouse)', default='GRCh38.mane.1.0.ensembl')
    parser.add_argument("--panel", help=HELP_DICT['panel'], choices=list(PANEL))
    parser.add_argument("--plot_top_n", type=int, help='plot UMAP of at most n variants ', default=20)
    if sub_program:
        s_common(parser)
        parser.add_argument('--match_dir', help=HELP_DICT['match_dir'], required=True)
        parser.add_argument('--vcf', help='vcf file.', required=True)



import subprocess
from collections import defaultdict
from multiprocessing import Pool
import math
import random

import pandas as pd
import numpy as np
import pysam

from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.flv_trust4.__init__ import CHAIN, REF_DIR, TOOLS_DIR

# split fastq into const N_CHUNK. Different N_CHUNK will cause different results as discussed in 
# https://github.com/liulab-dfci/TRUST4/issues/75
SPLIT_N_CHUNKS = 4


class Assemble(Step):
    """
    ## Features

    - TRUST4 does not use multi-processing when assembling. By default, the candidate reads are split into 4 chunks to speed up.

    - Keep only full-length contigs.

    ## Output
    - `03.assemble/assemble/{sample}_cdr3.out` All assembled CDR3 output and gene information.

    - `03.assemble/assemble/{sample}_assign.out` Read assignment results.

    - `03.assemble/assemble/{sample}_assembled_reads.fa` Assembled raw reads sequence.

    - `03.assemble/assemble/{sample}_annotate.fa` Assembled annotated contig sequences info.

    - `03.assemble/assemble/{sample}_report.tsv` Record assembled CDR3 types, read count and proportion.

    - `03.assemble/assemble/{sample}_filter_report.tsv` Filter non-functional CDR3.

    - `03.assemble/assemble/{sample}_barcode_report.tsv` Record chain information for each barcode.
    
    - `03.assemble/assemble/{sample}_barcode_filter_report.tsv` If --diffuseFrac provided. Filter low abundance barcode.
    """
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.ref = args.ref
        self.seqtype = args.seqtype
        self.barcodeRange = args.barcodeRange
        self.umiRange = args.umiRange
        self.candidate_fq = args.candidate_fq

        if args.not_split:
            self._n_chunk = 1
        else:
            self._n_chunk = SPLIT_N_CHUNKS
        self._chains = CHAIN[self.seqtype]
        # total thread may be higher than argument thread
        self._single_thread = math.ceil(self.thread / SPLIT_N_CHUNKS)

        # outdir
        self.assemble_outdir = f'{self.outdir}/assemble'
        self.temp_outdir = f'{self.assemble_outdir}/temp'
        for d in [self.assemble_outdir, self.temp_outdir]:
            utils.check_mkdir(dir_name=d)

        self.temp_outdir_list = [self.temp_outdir] * self._n_chunk
        self.temp_ref_list = [self.ref] * self._n_chunk
        self.temp_name_list = [f'temp_{i}' for i in range(self._n_chunk)]
        self.single_thread_list = [self._single_thread] * self._n_chunk

    def run(self):
        self.split_candidate_reads()
        self.run_assemble()
        self.run_annotate()
        self.merge_file()
        self.gen_report()

    @utils.add_log
    def split_candidate_reads(self):
        """
        split original candidate reads(_bcrtcr.fq) by barcode into N_CHUNK files
        """
        read_count_dict, umi_dict = defaultdict(int), defaultdict(set)
        with pysam.FastxFile(self.candidate_fq) as f:
            for read in f:
                attrs = read.name.split(':')
                cb, umi = attrs[0], attrs[1]
                read_count_dict[cb] += 1
                umi_dict[cb].add(umi)
        barcode_list = list(read_count_dict.keys())
        df_count = pd.DataFrame({'barcode': barcode_list, 
                            'read_count': [read_count_dict[i] for i in barcode_list], 
                            'UMI': [len(umi_dict[i]) for i in barcode_list]})
        df_count.sort_values(by='UMI', ascending=False, inplace=True)
        df_count.to_csv(f'{self.assemble_outdir}/count.txt', sep='\t', index=False)

        del read_count_dict
        del umi_dict

        # Split barcode list into equal size chunks.
        random.shuffle(barcode_list)
        split_barcodes = np.array_split(barcode_list, self._n_chunk)
        split_barcodes = [set(i) for i in split_barcodes]

        fq_list = [open(f'{self.temp_outdir}/temp_{i}.fq','w') for i in range(self._n_chunk)]
        bc_list = [open(f'{self.temp_outdir}/temp_{i}_bc.fa','w') for i in range(self._n_chunk)]
        umi_list = [open(f'{self.temp_outdir}/temp_{i}_umi.fa','w') for i in range(self._n_chunk)]

        with pysam.FastxFile(self.candidate_fq) as f:
            for read in f:
                name = read.name
                bc, umi = name.split(':')[0], name.split(':')[1]
                for i in range(self._n_chunk):
                    if bc in split_barcodes[i]:
                        fq_list[i].write(str(read) + '\n')
                        bc_list[i].write(f'>{name}\n{bc}\n')
                        umi_list[i].write(f'>{name}\n{umi}\n')
                        break

        for i in range(self._n_chunk):
            fq_list[i].close()
            bc_list[i].close()
            umi_list[i].close()
    
    @utils.add_log
    def run_assemble(self):
        """
        run assemble for each chunk
        """
        with Pool(self._n_chunk) as pool:
            pool.starmap(
                Assemble.assemble, 
                zip(self.temp_ref_list, self.temp_outdir_list, self.temp_name_list, self.single_thread_list)
            )

    @staticmethod
    @utils.add_log
    def assemble(ref, outdir, sample, single_thread, trimLevel=1):

        cmd = (
            f'trust4 -t {single_thread} '
            f'-f {REF_DIR}/{ref}/bcrtcr.fa '
            f'-o {outdir}/{sample} '
            f'-u {outdir}/{sample}.fq '
            f'--barcode {outdir}/{sample}_bc.fa '
            f'--UMI {outdir}/{sample}_umi.fa '
            f'--trimLevel {trimLevel} '
            '2>&1 '
        )
        Assemble.assemble.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run_annotate(self):
        """
        run annotate for each chunk
        """
        with Pool(self._n_chunk) as pool:
            pool.starmap(Assemble.annotate, zip(self.temp_name_list, self.temp_outdir_list, self.temp_ref_list, self.single_thread_list))

    @staticmethod
    @utils.add_log
    def annotate(name, outdir, ref, single_thread):

        cmd = (
            f'annotator -f {REF_DIR}/{ref}/IMGT+C.fa '
            f'-a {outdir}/{name}_final.out '
            f'-t {single_thread} '
            f'-o {outdir}/{name} '
            f'--barcode --UMI --noImpute '
            f'--readAssignment {outdir}/{name}_assign.out '
            f'-r {outdir}/{name}_assembled_reads.fa > {outdir}/{name}_annotate.fa '
        )
        Assemble.annotate.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def merge_file(self):
        """
        merge files from all threads into one file.
        """

        file_suffixes = [
            'annotate.fa',
            'cdr3.out',
            'assembled_reads.fa',
            'assign.out',
        ]

        for file_suffix in file_suffixes:
            file_path = [f'{self.temp_outdir}/temp_{thread}_{file_suffix}' for thread in range(self._n_chunk)]
            file_path_str = ' '.join(file_path)
            cmd = f'cat {file_path_str} > {self.assemble_outdir}/{self.sample}_{file_suffix}'
            self.merge_file.logger.info(cmd)
            self.debug_subprocess_call(cmd)

    def gen_report(self):
        Assemble.get_trust_report(self.assemble_outdir, self.sample)
        Assemble.filter_trust_report(self.assemble_outdir, self.sample)
        Assemble.get_bc_report(self.assemble_outdir, self.sample)
        Assemble.get_bcfilter_report(self.assemble_outdir, self.sample)
        Assemble.get_airr_file(self.assemble_outdir, self.sample)


    @staticmethod
    @utils.add_log
    def get_trust_report(filedir, sample):
        cmd = (
            f'perl {TOOLS_DIR}/trust-simplerep.pl '
            f'{filedir}/{sample}_cdr3.out > '
            f'{filedir}/{sample}_report.tsv '
            '2>&1 '
        )
        Assemble.get_trust_report.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @staticmethod
    @utils.add_log
    def filter_trust_report(filedir, sample):
        cmd = f''' awk '$4!~"_" && $4!~"?"' {filedir}/{sample}_report.tsv > {filedir}/{sample}_filter_report.tsv 2>&1'''
        Assemble.filter_trust_report.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @staticmethod
    @utils.add_log
    def get_bc_report(filedir, sample):
        cmd = (
            f'perl {TOOLS_DIR}/trust-barcoderep.pl '
            f'{filedir}/{sample}_cdr3.out > '
            f'{filedir}/{sample}_barcode_report.tsv '
            '2>&1 ' 
        )
        Assemble.get_bc_report.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @staticmethod
    @utils.add_log
    def get_bcfilter_report(filedir, sample):
        cmd = (
            f'python {TOOLS_DIR}/barcoderep-filter.py '
            f'-b {filedir}/{sample}_barcode_report.tsv > '
            f'{filedir}/{sample}_barcode_filter_report.tsv '
            '2>&1 '
        )
        Assemble.get_bcfilter_report.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)
    
    @staticmethod
    @utils.add_log
    def get_airr_file(filedir, sample):
        cmd = (
            f'perl {TOOLS_DIR}/trust-airr.pl '
            f'{filedir}/{sample}_barcode_report.tsv '
            f'{filedir}/{sample}_annotate.fa --format barcoderep > '
            f'{filedir}/airr_rearrangement.tsv '
            '2>&1 '
        )
        Assemble.get_airr_file.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)


@utils.add_log
def assemble(args):
    assemble_obj = Assemble(args)
    assemble_obj.run()


def get_opts_assemble(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--candidate_fq', help='Candidate fastq file from mapping step', required=True)

    parser.add_argument('--not_split', help='do not split reads into chunks', action='store_true')
    parser.add_argument('--ref', help='reference name', choices=["hg19", "hg38", "GRCm38", "other"], required=True)
    parser.add_argument('--seqtype', help='TCR/BCR seq data.', choices=['TCR', 'BCR'], required=True)
    parser.add_argument('--barcodeRange', help='Barcode range in fq1, INT INT CHAR.', default='0 23 +') 
    parser.add_argument('--umiRange', help='UMI range in fq1, INT INT CHAR.', default='24 -1 +')



import os
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor

import pandas as pd
import pysam

import celescope
from celescope.tools.utils import add_log, get_barcode_from_match_dir


@add_log
def razer(fq, outdir, sample, thread):

    # get ref
    root_path = os.path.dirname(celescope.__file__)
    ref = f'{root_path}/data/HLA/hla_reference_rna.fasta'

    # mkdir
    out_bam_dir = f'{outdir}/bam/'
    if not os.path.exists(out_bam_dir):
        os.mkdir(out_bam_dir)

    # out_bam
    out_bam = f'{outdir}/bam/{sample}.bam'

    # run
    cmd = (
        f'razers3 -i 97 -m 99999 --distance-range 0 -pa '
        f'-tc {thread} '
        f'-o {out_bam} '
        f'{ref} '
        f'{fq} '
    )
    razer.logger.info(cmd)
    os.system(cmd)
    return out_bam


@add_log
def split_bam(out_bam, barcodes, outdir, sample):
    '''
    input:
        out_bam: from razers3
        barcodes: cell barcodes
    ouput:
        bam_dict: assign reads to cell barcodes and UMI
        count_dict: UMI counts per cell
        index: assign index(1-based) to cells
    '''

    # init
    count_dict = defaultdict(dict)
    bam_dict = defaultdict(dict)
    index_dict = defaultdict(dict)
    cells_dir = f'{outdir}/cells/'

    # read bam and split
    split_bam.logger.info('reading bam...')
    samfile = pysam.AlignmentFile(out_bam, "rb")
    header = samfile.header
    for read in samfile:
        attr = read.query_name.split(':')
        barcode = attr[0]
        umi = attr[1]
        if barcode in barcodes:
            # keep one read for each UMI
            if umi not in bam_dict[barcode]:
                bam_dict[barcode][umi] = read
            if umi in count_dict[barcode]:
                count_dict[barcode][umi] += 1
            else:
                count_dict[barcode][umi] = 1

    split_bam.logger.info('writing cell bam...')
    # write new bam
    index = 0
    for barcode in barcodes:
        # init
        index += 1
        index_dict[index]['barcode'] = barcode
        index_dict[index]['valid'] = False

        # out
        if barcode in bam_dict:
            cell_dir = f'{cells_dir}/cell{index}'
            cell_bam_file = f'{cell_dir}/cell{index}.bam'
            if not os.path.exists(cell_dir):
                os.makedirs(cell_dir)
            index_dict[index]['valid'] = True
            cell_bam = pysam.AlignmentFile(
                f'{cell_bam_file}', "wb", header=header)
            for umi in bam_dict[barcode]:
                read = bam_dict[barcode][umi]
                cell_bam.write(read)
            cell_bam.close()

    # out df_index
    df_index = pd.DataFrame(index_dict).T
    df_index.index.name = 'cell_index'
    index_file = f'{outdir}/{sample}_cell_index.tsv'
    df_index.to_csv(index_file, sep='\t')

    # out count_dict
    df_count = pd.DataFrame(count_dict).T
    df_temp = df_count.stack()
    df_temp = df_temp.astype(int)
    df_temp = df_temp.reset_index()
    df_temp.columns = ['barcode', 'UMI', 'read_count']
    count_file = f'{outdir}/{sample}_UMI_count.tsv'
    df_temp.to_csv(count_file, sep='\t', index=False)

    return index_file, count_file


def sub_typing(bam):

    outdir = os.path.dirname(bam)
    prefix = os.path.basename(bam).strip('.bam')
    cmd = (
        f'OptiTypePipeline.py '
        f'--input {bam} '
        f'--rna '
        f'--outdir {outdir} '
        f'--prefix {prefix} '
        f'>/dev/null 2>&1 '
    )
    os.system(cmd)


def read_index(index_file):
    df_index = pd.read_csv(index_file, sep='\t', index_col=0, dtype=object)
    df_valid = df_index[df_index['valid'] == 'True']
    return df_valid


@add_log
def hla_typing(index_file, outdir, thread):
    all_res = []
    df_valid = read_index(index_file)

    bams = [
        f'{outdir}/cells/cell{index}/cell{index}.bam' for index in df_valid.index]
    with ProcessPoolExecutor(thread) as pool:
        for res in pool.map(sub_typing, bams):
            all_res.append(res)


@add_log
def summary(index_file, outdir, sample):

    n = 0
    df_valid = read_index(index_file)

    for index in df_valid.index:
        try:
            sub_df = pd.read_csv(
                f'{outdir}/cells/cell{index}/cell{index}_result.tsv', sep='\t', index_col=0)
        except FileNotFoundError:
            continue
        n += 1
        sub_df['barcode'] = df_valid.loc[index, :]['barcode']
        sub_df['cell_index'] = index
        if n == 1:
            all_df = sub_df
        else:
            all_df = all_df.append(sub_df, ignore_index=True)
    all_df['Reads'] = all_df['Reads'].apply(int)
    all_df = all_df[all_df['Reads'] != 0]
    all_df = all_df.drop('Objective', axis=1)
    out_file = f'{outdir}/{sample}_typing.tsv'
    all_df.to_csv(out_file, sep='\t', index=False)


@add_log
def mapping_hla(args):

    sample = args.sample
    outdir = args.outdir
    fq = args.fq
    thread = int(args.thread)
    match_dir = args.match_dir

    # process args
    barcodes, _nCell = get_barcode_from_match_dir(match_dir)

    # check dir
    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % (outdir))

    # razer
    out_bam = razer(fq, outdir, sample, thread)

    # split bam
    index_file, _count_file = split_bam(out_bam, barcodes, outdir, sample)

    # typing
    hla_typing(index_file, outdir, thread)

    # summary
    summary(index_file, outdir, sample)


def get_opts_mapping_hla(parser, sub_program):
    if sub_program:
        parser.add_argument('--outdir', help='output dir', required=True)
        parser.add_argument('--sample', help='sample name', required=True)
        parser.add_argument("--fq", required=True)
        parser.add_argument('--assay', help='assay', required=True)
    parser.add_argument(
        "--match_dir", help="match scRNA-Seq dir", required=True)
    parser.add_argument("--thread", help='number of thread', default=1)



from celescope.tools import utils
from celescope.tools.star_mixin import Star_mixin, get_opts_star_mixin


class Star_fusion(Star_mixin):
    """
    ## Features
    - The reads were aligned to the known fusion sites(specified by `--fusion_genomeDir`) using STAR. 
    Please note that [STAR-Fusion](https://github.com/STAR-Fusion/STAR-Fusion/wiki) is not used here.
    """
    def __init__(self, args, display_title=None):
        args.genomeDir = args.fusion_genomeDir
        super().__init__(args, display_title)


@utils.add_log
def star_fusion(args):
    with Star_fusion(args) as runner:
        runner.run()


def get_opts_star_fusion(parser, sub_program):
    get_opts_star_mixin(parser, sub_program)
    # will cause `conflicting option string: --genomeDir`
    # parser.add_argument('--genomeDir', help=argparse.SUPPRESS)
    parser.add_argument('--fusion_genomeDir', help='fusion gene STAR index genome directory', required=True)



import numpy as np
import scipy.stats as sp_stats


class SimpleGoodTuringError(Exception):
    pass


def _averaging_transform(r, nr):
    d = np.concatenate((np.ones(1, dtype=int), np.diff(r)))
    dr = np.concatenate((
        0.5 * (d[1:] + d[0:-1]),
        np.array((d[-1],), dtype=float),
    ))
    return nr.astype(float)/dr


def _rstest(r, coef):
    return r * np.power(1 + 1/r, 1 + coef)


def simple_good_turing(xr, xnr):
    """Make a Simple Good-Turing estimate of the frequencies.
    Args:
      xr (np.array(int)): Non-zero item frequencies
      xnr (np.array(int)): Non-zero frequencies of frequencies
    Returns:
      (rstar (np.array(float)), p0 (float)):
        rstar: The adjusted non-zero frequencies
        p0: The total probability of unobserved items
    """

    xr = xr.astype(float)
    xnr = xnr.astype(float)

    xN = np.sum(xr*xnr)

    # Get Linear Good-Turing estimate
    xnrz = _averaging_transform(xr, xnr)
    slope, _intercept, _, _, _ = sp_stats.linregress(np.log(xr), np.log(xnrz))

    if slope > -1:
        raise SimpleGoodTuringError(
            "The log-log slope is > -1 (%d); the SGT estimator is not applicable to these data." % slope)

    xrst = _rstest(xr, slope)
    xrstrel = xrst/xr

    # Get traditional Good-Turing estimate
    xrtry = xr == np.concatenate((xr[1:]-1, np.zeros(1)))
    xrstarel = np.zeros(len(xr))
    xrstarel[xrtry] = (xr[xrtry]+1) / xr[xrtry] * np.concatenate((xnr[1:], np.zeros(1)))[xrtry] / xnr[xrtry]

    # Determine when to switch from GT to LGT estimates
    tursd = np.ones(len(xr))
    for i in range(len(xr)):
        if xrtry[i]:
            tursd[i] = float(i+2) / xnr[i] * np.sqrt(xnr[i+1] * (1 + xnr[i+1]/xnr[i]))

    xrstcmbrel = np.zeros(len(xr))
    useturing = True
    for r in range(len(xr)):
        if not useturing:
            xrstcmbrel[r] = xrstrel[r]
        else:
            if np.abs(xrstrel[r]-xrstarel[r]) * (1+r)/tursd[r] > 1.65:
                xrstcmbrel[r] = xrstarel[r]
            else:
                useturing = False
                xrstcmbrel[r] = xrstrel[r]

    # Renormalize the probabilities for observed objects
    sumpraw = np.sum(xrstcmbrel * xr * xnr / xN)

    xrstcmbrel = xrstcmbrel * (1 - xnr[0] / xN) / sumpraw
    p0 = xnr[0]/xN

    return (xr * xrstcmbrel, p0)


def sgt_proportions(frequencies):
    """Use Simple Good-Turing estimate to adjust for unobserved items
    Args:
      frequencies (np.array(int)): Nonzero frequencies of items
    Returns:
      (pstar (np.array(float)), p0 (float)):
        pstar: The adjusted non-zero proportions
        p0: The total probability of unobserved items
    """
    if len(frequencies) == 0:
        raise ValueError("Input frequency vector is empty")
    if np.count_nonzero(frequencies) != len(frequencies):
        raise ValueError("Frequencies must be greater than zero")

    freqfreqs = np.bincount(frequencies)
    assert freqfreqs[0] == 0
    use_freqs = np.flatnonzero(freqfreqs)

    if len(use_freqs) < 10:
        raise SimpleGoodTuringError("Too few non-zero frequency items (%d). Aborting SGT." % len(use_freqs))

    rstar, p0 = simple_good_turing(use_freqs, freqfreqs[use_freqs])

    # rstar contains the smoothed frequencies.
    # Map each original frequency r to its smoothed rstar.
    rstar_dict = dict(zip(use_freqs, rstar))

    rstar_sum = np.sum(freqfreqs[use_freqs] * rstar)
    rstar_i = np.fromiter((rstar_dict[f] for f in frequencies),
                          dtype=float, count=len(frequencies))
    pstar = (1 - p0) * (rstar_i / rstar_sum)

    assert np.isclose(p0 + np.sum(pstar), 1)
    return (pstar, p0)



from celescope.tools.tag.mapping_tag import Mapping_tag as Mt, get_opts_mapping_tag as opts, mapping_tag as mt

class Mapping_tag(Mt):
    pass

def mapping_tag(args):
    mt(args)

def get_opts_mapping_tag(parser, sub_program):
    opts(parser, sub_program)


from celescope.flv_trust4.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_flv_trust4(Multi):
    """

    ## Usage
    
    ```
    multi_flv_trust4 \\
        --mapfile ./test.mapfile \\
        --ref hg38 \\
        --thread 8 \\
        --seqtype BCR \\
        --mod shell
    ```
    """
    def barcode(self, sample):
        step = "barcode"
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr[0]} --fq2 {arr[1]} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def mapping(self, sample):
        step = 'mapping'
        cmd_line = self.get_cmd_line(step, sample)
        match_fq1 = f'{self.outdir_dic[sample]["barcode"]}/{sample}_1.fq'
        match_fq2 = f'{self.outdir_dic[sample]["barcode"]}/{sample}_2.fq'
        cmd = (
            f'{cmd_line} '
            f'--match_fq1 {match_fq1} '
            f'--match_fq2 {match_fq2} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=self.args.thread)

    def assemble(self, sample):
        step = 'assemble'
        cmd_line = self.get_cmd_line(step, sample)
        candidate_fq = f'{self.outdir_dic[sample]["mapping"]}/{sample}_bcrtcr.fq'
        cmd = (
            f'{cmd_line} '
            f'--candidate_fq {candidate_fq} '
        )
        self.process_cmd(cmd, step, sample, m=30, x=self.args.thread)
    
    def summarize(self, sample):
        step = 'summarize'
        cmd_line = self.get_cmd_line(step, sample)
        assemble_out = f'{self.outdir_dic[sample]["assemble"]}/assemble'
        fq2 = f'{self.outdir_dic[sample]["barcode"]}/{sample}_2.fq'
        cmd = (
            f'{cmd_line} '
            f'--assemble_out {assemble_out} '
            f'--fq2 {fq2} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=15, x=1)

    def annotation(self,sample):
        step = 'annotation'
        cmd_line = self.get_cmd_line(step,sample)
        summarize_out = f'{self.outdir_dic[sample]["summarize"]}'
        match_dir = f'{self.col4_dict[sample]}'
        cmd = (
            f'{cmd_line} '
            f'--summarize_out {summarize_out} '
            f'--match_dir {match_dir} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)


def main():
    multi = Multi_flv_trust4(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()


import collections
import math

import numpy as np
import pandas as pd
import plotly.graph_objs as go
import plotly.offline as pltoff

from celescope.tools.utils import add_log

BarcodeRankPlotSegment = collections.namedtuple('BarcodeRankPlotSegment', ['start', 'end', 'cell_density', 'legend'])

CHARTS_PLOTLY_MODEBAR_TRANSFORM_BUTTONS = [
    'zoom2d',
    'pan2d',
    'zoomIn2d',
    'zoomOut2d',
    'autoScale2d',
    # 'resetScale2d'  can't totally disable interaction, it seems-- keep reset option
]

CHARTS_PLOTLY_EXPORT_BUTTONS = [
    'toImage',
    'sendDataToCloud',
]

CHARTS_PLOTLY_FIXED_CONFIG = {
    'modeBarButtonsToRemove': CHARTS_PLOTLY_MODEBAR_TRANSFORM_BUTTONS+CHARTS_PLOTLY_EXPORT_BUTTONS,
    'displaylogo': False,
    'showLink': False
}

CHARTS_PLOTLY_MOVABLE_CONFIG = {
    'modeBarButtonsToRemove': CHARTS_PLOTLY_EXPORT_BUTTONS,
    'displaylogo': False,
    'showLink': False
}
BC_RANK_PLOT_LINE_WIDTH = 3
# Gradient scheme used in the barcode rank plot
BC_PLOT_COLORS = ['#dddddd', '#d1d8dc', '#c6d3dc', '#bacfdb', '#aecada', '#a3c5d9', '#97c0d9', '#8cbbd8', '#80b7d7',
                  '#74b2d7', '#6aadd6', '#66abd4', '#62a8d2', '#5ea5d1', '#59a2cf', '#559fce', '#519ccc', '#4d99ca',
                  '#4997c9', '#4594c7', '#4191c5', '#3d8dc4', '#3a8ac2', '#3787c0', '#3383be', '#3080bd', '#2c7cbb',
                  '#2979b9', '#2676b7', '#2272b6', '#1f6eb3', '#1d6ab0', '#1a65ac', '#1861a9', '#155ca6', '#1358a2',
                  '#10539f', '#0e4f9b', '#0b4a98', '#094695', '#09438f', '#0a4189', '#0c3f83', '#0d3d7c', '#0e3b76',
                  '#103970', '#11366a', '#123463', '#14325d', '#153057']

CHARTS = [
    {
        'layout': {
            'title': 'Barcode Rank',
            'width': 470,
            'height': 313,
            'margin': {'l': 60, 'r': 0, 't': 30, 'b': 40},
            'hovermode': 'closest',
            'xaxis': {
                'title': 'Barcodes',
                'type': 'log',
            },
            'yaxis': {
                'title': 'UMI counts',
                'type': 'log',
            },
        },
        'data': [],
        'config': CHARTS_PLOTLY_FIXED_CONFIG,
        'function': '_plot_barcode_rank',
        'description': 'The plot shows the count of filtered UMIs mapped to each barcode.  Barcodes can be determined to be cell-associated based on their UMI counts or by their expression profiles. Therefore some regions of the graph can contain both cell-associated and background-associated barcodes. The color of the graph represents the local density of barcodes that are cell-associated.',
        'name': 'barcode_rank',
    },
]


def BC_PLOT_CMAP(density):
    """
    Colormap utility fn to map a number to one of the colors in the gradient
    color scheme defined above
    Input
    - density : A real number in the range [0,1]
    """
    assert density >= 0.
    assert density <= 1.
    levels = len(BC_PLOT_COLORS)
    ind = min(levels - 1, int(math.floor(levels * density)))
    return BC_PLOT_COLORS[ind]


def get_plot_segment(start_index, end_index, sorted_bc, cell_barcodes, legend=False):
    """
    Helper function to build a plot segment.
    """
    assert end_index > start_index
    num_cells = sum([1 for i in range(start_index, end_index) if sorted_bc[i] in cell_barcodes])
    density = float(num_cells)/float(end_index-start_index)
    return BarcodeRankPlotSegment(start=start_index, end=end_index, cell_density=density, legend=legend)


def segment_log_plot_by_length(y_data, x_start, x_end):
    """
    Given the extends of the mixed region [x_start, x_end), compute
    the x-indices that would divide the plot into segments of a
    prescribed length (in pixel coordinates) with a minimum number
    of barcodes in each segment
    """

    if x_end <= x_start:
        return []

    SEGMENT_NORMALIZED_MAX_LEN = 0.02
    MIN_X_SPAN = 20

    log_max_x = np.log(len(y_data))
    log_max_y = np.log(max(y_data))

    this_segment_len = 0.0
    segment_idx = [x_start]

    for i in range(x_start, x_end):
        last_i = max(x_start, i-1)
        dx = (np.log(i) - np.log(last_i)) / log_max_x
        dy = (np.log(y_data[i]) - np.log(y_data[last_i])) / log_max_y
        this_segment_len += np.linalg.norm([dx, dy])
        if this_segment_len >= SEGMENT_NORMALIZED_MAX_LEN and i > (segment_idx[-1] + MIN_X_SPAN):
            segment_idx.append(i+1)
            this_segment_len = 0.0

    if segment_idx[-1] != x_end:
        segment_idx.append(x_end)

    return segment_idx


def segment_log_plot_by_length_log_uniform(x_start, x_end):
    """
    vdj usually have very large x_end
    """

    if x_end <= x_start:
        return []

    MIN_X_LOG10_SPAN = 0.03
    MIN_BARCODE = 20
    MAX_N = 100
    segment_idx = [x_start]
    span = np.log10(x_end-x_start) / MAX_N
    if span < MIN_X_LOG10_SPAN:
        span = MIN_X_LOG10_SPAN
        
    span_sum = 1
    idx = x_start
    while True:
        span_sum += span
        idx = int(x_start + 10 ** span_sum)
        if idx - segment_idx[-1] < MIN_BARCODE:
            continue
        if idx > x_end:
            break
        segment_idx.append(idx)

    if segment_idx[-1] != x_end:
        segment_idx.append(x_end)

    return segment_idx


def convert_numpy_array_to_line_chart(array, ntype):
    array = np.sort(array)[::-1]

    rows = []
    previous_count = None
    for (index,), count in np.ndenumerate(array):
        if index == 0 or index == len(array)-1:
            rows.append([index, ntype(count)])
        elif previous_count != count:
            previous_index = rows[-1][0]
            if previous_index != index - 1:
                rows.append([index - 1, ntype(previous_count)])
            rows.append([index, ntype(count)])
        previous_count = count
    return rows


@add_log
def counter_barcode_rank_plot_data(df_UMI_path, log_uniform=False):
    """
    get cell density for each plot_segments
    :param count_data_path:
    :return: sorted_counts, plot_segments, cell_nums
    """
    df_UMI = pd.read_csv(df_UMI_path, index_col=0, sep='\t')
    cell_bc = set(df_UMI[df_UMI['mark'] == 'CB'].index)
    sorted_bc = np.array(df_UMI.index)
    sorted_counts = np.array(df_UMI['UMI'])
    cell_nums = len(cell_bc)
    total_bc = len(sorted_bc)
    # find the first barcode which is not a cell
    first_non_cell = total_bc
    for i, bc in enumerate(sorted_bc):
        if bc not in cell_bc:
            first_non_cell = i
            break

    # find the last barcode which is a cell
    last_cell = 0
    for i in reversed(range(total_bc)):
        if sorted_bc[i] in cell_bc:
            last_cell = i
            break

    ranges = [0, first_non_cell, last_cell+1, total_bc]
    plot_segments = []
    plot_segments.append(BarcodeRankPlotSegment(start=0, end=ranges[1], cell_density=1.0, legend=True))
    plot_segments.append(BarcodeRankPlotSegment(start=ranges[2], end=ranges[3], cell_density=0.0, legend=True))

    if not log_uniform:
        mixed_segments = segment_log_plot_by_length(sorted_counts, ranges[1], ranges[2])
    else:
        mixed_segments = segment_log_plot_by_length_log_uniform(ranges[1], ranges[2])
    for i in range(len(mixed_segments) - 1):
        plot_segments.append(
            get_plot_segment(mixed_segments[i], mixed_segments[i + 1], sorted_bc, cell_bc, legend=False))

    return sorted_counts, plot_segments, cell_nums


def _plot_barcode_rank(chart, counts, num_cells):
    """ Generate a generic barcode rank plot """
    rows = convert_numpy_array_to_line_chart(counts, int)

    for _i, row in enumerate(rows):
        index, count = row[0], row[1]
        if index < num_cells:
            series_list = [chart['data'][0]]
        elif index == num_cells:
            # Connect the two lines
            series_list = [chart['data'][0], chart['data'][1]]
        else:
            series_list = [chart['data'][1]]

        for series in series_list:
            series['x'].append(index)
            series['y'].append(count)

    # Handle case where there is no background
    bg_series = chart['data'][1]
    if len(bg_series['x']) == 0:
        bg_series['x'].append(0)
        bg_series['y'].append(0)

    return chart


def build_plot_data_dict(plot_segment, counts):
    """
    Construct the data for a plot segment by appropriately slicing the
    counts
    Inputs:
    - plot_segment: BarcodeRankPlotSegment containing [start, end)
        of the segment, the cell density and legend visibility option
    - counts: Reverse sorted UMI counts for all barcodes.
    """

    start = max(0, plot_segment.start - 1)  # -1 for continuity between two charts
    end = plot_segment.end
    plot_rows = convert_numpy_array_to_line_chart(counts[start:end], int)
    name = 'Cells' if plot_segment.cell_density > 0 else 'Background'

    # Setup the tooltip
    if plot_segment.cell_density > 0.:
        n_barcodes = plot_segment.end - plot_segment.start
        n_cells = int(round(plot_segment.cell_density * n_barcodes))
        hover = "{:.0f}% Cells<br>({}/{})".format(100 * plot_segment.cell_density, n_cells, n_barcodes)
    else:
        hover = "Background"

    data_dict = {
        "x": [],
        "y": [],
        "name": name,
        "hoverinfo": "text",
        "text": hover,
        "type": "scattergl",
        "mode": "lines",
        "line": {
            "width": 3,
            "color": BC_PLOT_CMAP(plot_segment.cell_density),
        },
        "showlegend": plot_segment.legend,
    }
    offset = 1 + start  # it's a log-log plot, hence the 1
    for index, count in plot_rows:
        data_dict["x"].append(index + offset)
        data_dict["y"].append(count)

    # Handle case where the data is empty
    if len(data_dict["x"]) == 0:
        data_dict["x"].append(0)
        data_dict["y"].append(0)

    return data_dict


def _plot_counter_barcode_rank(chart, counts, plot_segments):
    """
    Generate the RNA counter barcode rank plot
    Inputs:
        - chart: chart element to populate data
        - counts: UMI counts reverse sorted
        - plot_segments: A list of BarcodeRankPlotSegments
    """

    for segment in plot_segments:
        chart["data"].append(build_plot_data_dict(segment, counts))

    return chart


@add_log
def get_plot_data(plot_segments, counts):
    plot_data = []
    for segment in plot_segments:
        plot_data.append(build_plot_data_dict(segment, counts))

    return plot_data


def plot_barcode_rank(count_file_path, log_uniform=False):
    sorted_counts, plot_segments, _cell_nums = counter_barcode_rank_plot_data(count_file_path, log_uniform=log_uniform)
    plot_data = get_plot_data(plot_segments, sorted_counts)

    plotly_data = [go.Scatter(x=dat['x'], y=dat['y'], name=dat['name'], mode=dat['mode'], showlegend=dat['showlegend'],
                              marker={'color': dat['line']['color']}, line=dat['line'], text=dat['text']) for dat in
                   plot_data]

    layout = go.Layout(width=470, height=313,
                       title={"text": "Barcode rank", "font": {"color": "black"}, "x": 0.5},
                       xaxis={"type": "log", "title": "Barcode", "titlefont": {"color": "black"},
                              "color": "black", "gridcolor": "gainsboro", "linecolor": "black"},
                       yaxis={"type": "log", "title": "UMI counts", "titlefont": {"color": "black"},
                              "color": "black", "gridcolor": "gainsboro", "linecolor": "black"},
                       margin=dict(l=50, r=0, t=30, b=30),
                       plot_bgcolor="#FFFFFF")

    config = dict({"displayModeBar": True,
                   "staticPlot": False,
                   "showAxisDragHandles": False,
                   "modeBarButtons": [["toImage", "resetScale2d"]],
                   "scrollZoom": False,
                   "displaylogo": False})

    fig = go.Figure(data=plotly_data, layout=layout)

    chart = pltoff.plot(fig, include_plotlyjs=False, output_type='div', config=config)

    return chart



import pandas as pd

from celescope.flv_trust4.summarize import Summarize
from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.tools.plotly_plot import Bar_plot
from celescope.flv_trust4.__init__ import CHAIN, PAIRED_CHAIN
from collections import OrderedDict


def gen_vj_annotation_metrics(df, seqtype):
    """
    Generate vdj Annotation Metrics from contig annotations file.
    """
    def get_vj_spanning_pair():
        """
        Get Productive V-J Spanning_Pair metric from annotation file
        Return productive chain pair number. eg: TRA/TRB or IGH/IGL, IGH/IGK.
        """
        df_productive = df[df['productive'] == True]
        
        if seqtype == "BCR":
            df_chain_heavy = df_productive[(df_productive['chain'] == 'IGH')] 
            df_chain_light = df_productive[(df_productive['chain'] == 'IGL') | (df_productive['chain'] =='IGK')]
        else:
            df_chain_heavy = df_productive[df_productive['chain'] == 'TRA']
            df_chain_light = df_productive[df_productive['chain'] == 'TRB']

        for _df in [df_chain_heavy, df_chain_light]:
            _df.drop_duplicates(['barcode'], inplace=True)

        VJ_Spanning_Pair_Cells = pd.merge(df_chain_heavy, df_chain_light, on='barcode', how='inner')
        
        return VJ_Spanning_Pair_Cells.shape[0]

    metric_dict = OrderedDict()
    chains, chain_pairs = CHAIN[seqtype], PAIRED_CHAIN[seqtype]

    metric_dict["Cells match with scRNA-seq analysis"] = len(set(df.barcode))
    metric_dict["Cells With Productive V-J Spanning Pair"] = get_vj_spanning_pair()

    for pair in chain_pairs:
        chain1, chain2 = pair.split('_')[0], pair.split('_')[1]
        cbs1 = set(df[(df['productive']==True)&(df['chain']==chain1)].barcode)
        cbs2 = set(df[(df['productive']==True)&(df['chain']==chain2)].barcode)
        metric_dict[f"Cells With Productive V-J Spanning ({chain1}, {chain2}) Pair"] = len(cbs1.intersection(cbs2))

    for chain in chains:
        metric_dict[f"Cells With {chain} Contig"] = len(set(df[df['chain']==chain].barcode))
        metric_dict[f"Cells With CDR3-annotated {chain} Contig"] = len(set(df[(df['chain']==chain)&(df['cdr3']!='None')].barcode))
        metric_dict[f"Cells With V-J Spanning {chain} Contig"] = len(set(df[(df['full_length']==True)&(df['chain']==chain)].barcode))
        metric_dict[f"Cells With Productive {chain} Contig"] =len(set(df[(df['productive']==True)&(df['chain'] == chain)].barcode))

    return metric_dict


def gen_clonotypes_table(df, out_clonotypes, seqtype):
    """
    Generate clonotypes.csv file
    """
    df = df[df['productive'] == True]
    df['chain_cdr3aa'] = df[['chain', 'cdr3']].apply(':'.join, axis=1)
    df = df.rename(columns={"chain_cdr3aa":"cdr3s_aa", "raw_clonotype_id":"clonotype_id"})
    df = df.dropna(subset=["clonotype_id"])
    df = df.sort_values("clonotype_id", key=lambda x: x.str.lstrip("clonotype").astype(int))
    
    sort_method = {"TCR": True, "BCR": False}
    cdr3_aa_dict = df.groupby("clonotype_id")["cdr3s_aa"].apply(set).to_dict()
    cdr3_aa_dict = {key: ';'.join(sorted(list(value), reverse=sort_method[seqtype])) for key, value in cdr3_aa_dict.items()}
    count_dict = df.groupby("clonotype_id")["barcode"].nunique().to_dict()
    
    df["frequency"] = df["clonotype_id"].apply(lambda x: count_dict[x])
    df["cdr3s_aa"] = df["clonotype_id"].apply(lambda x: cdr3_aa_dict[x])
    df = df.drop_duplicates("clonotype_id")
    df = df[["clonotype_id", "cdr3s_aa", "frequency"]]
    
    sum_frequency = df['frequency'].sum()
    df['proportion'] = df['frequency'].apply(lambda x: x/sum_frequency)

    df.to_csv(out_clonotypes, sep=',', index=False)
    

class Annotation(Step):
    """
    ## Features

    - Output TSNE-plot of Assembled T/B Cells.

    ## Output
    - `05.annotation/{sample}_mapping.pdf` TSNE-plot of Assembled Cells.

    """
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.match_dir = args.match_dir
        self.chains, self.paired_groups = Summarize._parse_seqtype(self.seqtype)
        self.contig_file = f'{args.summarize_out}/{self.sample}_filtered_contig.csv'
        self.clonotype_file = f'{args.summarize_out}/clonotypes.csv'        


    def parse_clonotype(self):
        """Generate clonotypes table in html.
        """
        df_clonotypes=pd.read_csv(self.clonotype_file, sep=',')
        df_clonotypes['ClonotypeID'] = df_clonotypes['clonotype_id'].apply(lambda x: x.strip('clonetype'))
        df_clonotypes['Frequency'] = df_clonotypes['frequency']
        df_clonotypes['Proportion'] = df_clonotypes['proportion'].apply(lambda x: f'{round(x*100, 2)}%')
        df_clonotypes['CDR3_aa'] = df_clonotypes['cdr3s_aa'].apply(lambda x: x.replace(';', '<br>'))

        return df_clonotypes

    @staticmethod
    @utils.add_log
    def get_vdj_metric(df, chains, pairs):
        """
        Add vdj metrics in html.
        """
        metric_result = []
        fl_pro_pair_df = pd.DataFrame(df[df['productive']==True].barcode.value_counts())
        fl_pro_pair_df = fl_pro_pair_df[fl_pro_pair_df['barcode']>=2]
        cell_nums = len(set(df['barcode']))

        metric_result.append({
            'name': 'Cells With Productive V-J Spanning Pair',
            'value': fl_pro_pair_df.shape[0],
            'total': cell_nums,
        })

        for pair in pairs:
            chain1, chain2 = pair.split('_')[0], pair.split('_')[1]
            cbs1 = set(df[(df['full_length']==True)&(df['productive']==True)&(df['chain']==chain1)].barcode)
            cbs2 = set(df[(df['full_length']==True)&(df['productive']==True)&(df['chain']==chain2)].barcode)
            paired_cbs = len(cbs1.intersection(cbs2))

            metric_result.append({
                'name': f'Cells With Productive V-J Spanning ({chain1}, {chain2}) Pair',
                'value': paired_cbs,
                'total': cell_nums,
                'help_info': "Fraction of cell-associated barcodes with one productive contig for each chain of the receptor pair.A productive contig satisfies the following conditions: the contig annotations span the 5' end of the V region to the 3' end of the J region of the chain, a start codon was found in the expected part of the V sequence, an in-frame CDR3 amino acid motif was found, and no stop codons were found in the aligned V-J region"
            })

        for chain in chains:
        
            metric_result.append({
                'name': f'Cells With {chain} Contig',
                'value': len(set(df[df['chain']==chain].barcode)),
                'total': cell_nums,
                'help_info': f'Fraction of cell-associated barcodes with at least one {chain} contig annotated as a full or partial V(D)J gene'
            })
            metric_result.append({
                'name': f'Cells With CDR3-annotated {chain} Contig',
                'value': len(set(df[(df['chain']==chain)&(df['cdr3']!=None)].barcode)),
                'total': cell_nums,
            })
            metric_result.append({
                'name': f'Cells With V-J Spanning {chain} Contig',
                'value': len(set(df[(df['full_length']==True)&(df['chain']==chain)].barcode)),
                'total': cell_nums,
                'help_info': f"Fraction of cell-associated barcodes with at least one contig spanning the 5' end of the V region to the 3' end of the J region for {chain}"
            })
            metric_result.append({
                'name': f'Cells With Productive {chain} Contig',
                'value': len(set(df[(df['full_length']==True)&(df['productive']==True)&(df['chain']==chain)].barcode)),
                'total': cell_nums,
                'help_info': "Fraction of cell-associated barcodes with productive IGL chain. A productive contig satisfies the following conditions: the contig annotations span the 5' end of the V region to the 3' end of the J region of the chain, a start codon was found in the expected part of the V sequence, an in-frame CDR3 amino acid motif was found, and no stop codons were found in the aligned V-J region"
            })

        return metric_result 

    @utils.add_log
    def annotation_process(self):
        """Add metrics, clonotypes table, and bar-plot of clonotypes distribution in html.
        """
        df = pd.read_csv(self.contig_file)

        annotation_summary = Annotation.get_vdj_metric(df, self.chains, self.paired_groups)
        for anno in annotation_summary:
            self.add_metric(anno['name'], anno['value'], anno.get('total'), anno.get('help_info'))

        df_clonotypes = self.parse_clonotype()
        title = 'Clonetypes'
        table_dict = self.get_table_dict(
            title = title,
            table_id = 'clonetypes',
            df_table = df_clonotypes[['ClonotypeID', 'CDR3_aa', 'Frequency', 'Proportion']]
        )
        self.add_data(table_dict=table_dict)

        df_clonotypes['ClonotypeID'] = df_clonotypes['ClonotypeID'].astype("int")
        df_clonotypes.sort_values(by=['ClonotypeID'], inplace=True)
        Barplot = Bar_plot(df_bar=df_clonotypes).get_plotly_div()
        self.add_data(Barplot=Barplot)
    
    def run(self):
        self.annotation_process()
        

@utils.add_log
def annotation(args):
    with Annotation(args, display_title="V(D)J Annotation") as runner:
        runner.run()


def get_opts_annotation(parser, sub_program):
    parser.add_argument('--seqtype', help='TCR or BCR',
                        choices=['TCR', 'BCR'], required=True)               
    if sub_program:
        s_common(parser)
        parser.add_argument('--match_dir', help='scRNA-seq match directory', required=True)
        parser.add_argument('--summarize_out', help='summarize output directory', required=True)  
    return parser


import pysam

from celescope.tools import utils
from celescope.tools.capture.threshold import Threshold
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT

# variant allele frequency (VAF) threshold
VAF = 0.2


class Filter_snp(Step):
    """
    ## Features
    - Filter out `ref` and `alt` alleles that do not have enough reads to support.

    ## Output
    - `{sample}_test1_filtered.vcf` VCF file after filtering. Alleles read counts that do not have enough reads to support are set to zero. 
    Genotypes are changed accordingly.
    """
    def __init__(self, args, display_title='Filtering'):
        super().__init__(args, display_title)
        self.vcf = args.vcf

        self.add_metric(
            'reference allele threshold method',
            self.args.ref_threshold_method,
        )

        self.add_metric(
            'alternate allele threshold method',
            self.args.alt_threshold_method,
        )

        # out
        self.out_vcf_file = f'{self.out_prefix}_filtered.vcf'


    @staticmethod
    def get_count_array(record):
        """
        get allele count array

        Args
            record: pysam vcf record
            entry: str, 'ref' or 'alt'
        Return
            (ref_count_array, alt_count_array)
        """
        ref_count_array = []
        alt_count_array = []
        for sample in record.samples:
            ad = record.samples[sample]['AD']
            ref_count_array.append(ad[0])
            alt_count_array.append(ad[1])

        return ref_count_array, alt_count_array

    def get_threshold(self, count_array, threshold_method, otsu_plot_path=None):
        runner = Threshold(
            count_array,
            threshold_method=threshold_method,
            otsu_plot_path=otsu_plot_path
        )
        threshold = runner.run()
        return threshold

    @staticmethod
    def filter_array(count_array, threshold):
        return [x if x >= threshold else 0 for x in count_array]

    @staticmethod
    def get_genotype(ref_count, alt_count):
        if ref_count > 0 and alt_count > 0:
            af = alt_count / (alt_count + ref_count)
            if VAF <= af <= 1 - VAF:
                return (0,1)
            elif af < VAF:
                return (0,0)
            else:
                return (1,1)
        elif ref_count > 0 and alt_count == 0:
            return (0,0)
        elif ref_count == 0 and alt_count > 0:
            return (1,1)
        elif ref_count == 0 and alt_count == 0:
            return (None, None)

    @utils.add_log
    def run(self):
        
        if self.args.ref_threshold_method == 'otsu':
            ref_otsu_plot_dir = f'{self.outdir}/ref_otsu_plots/'
            utils.check_mkdir(ref_otsu_plot_dir)
        
        if self.args.alt_threshold_method == 'otsu':
            alt_otsu_plot_dir = f'{self.outdir}/alt_otsu_plots/'
            utils.check_mkdir(alt_otsu_plot_dir)

        with pysam.VariantFile(self.vcf) as vcf_in:
            header = vcf_in.header
            header.add_meta('ref_threshold_method', value=self.args.ref_threshold_method)
            header.add_meta('alt_threshold_method', value=self.args.alt_threshold_method)
            header.add_meta('INFO', items=[('ID',"REF_T"), ('Number',1), ('Type','Integer'), ('Description','Reference allele count threshold')])
            header.add_meta('INFO', items=[('ID',"ALT_T"), ('Number',1), ('Type','Integer'), ('Description','Alternate allele count threshold')])
            with pysam.VariantFile(self.out_vcf_file, 'w', header=vcf_in.header) as vcf_out:
                for record in vcf_in.fetch():
                    name = f'{record.chrom}_{record.pos}'
                    ref_count_array, alt_count_array = self.get_count_array(record)
                    ref_otsu_plot_path = f'{ref_otsu_plot_dir}/{name}_ref.pdf' if self.args.ref_threshold_method == 'otsu' else None
                    ref_threshold = self.get_threshold(ref_count_array, self.args.ref_threshold_method, ref_otsu_plot_path)
                    alt_otsu_plot_path = f'{alt_otsu_plot_dir}/{name}_alt.pdf' if self.args.alt_threshold_method == 'otsu' else None
                    alt_threshold = self.get_threshold(alt_count_array, self.args.alt_threshold_method, alt_otsu_plot_path)
                    ref_threshold = max(ref_threshold, self.args.ref_min_support_read)
                    alt_threshold = max(alt_threshold, self.args.alt_min_support_read)
                    ref_filtered_count_array = self.filter_array(ref_count_array, ref_threshold)
                    alt_filtered_count_array = self.filter_array(alt_count_array, alt_threshold)

                    new_record = record.copy()
                    new_record.info.__setitem__('REF_T', ref_threshold)
                    new_record.info.__setitem__('ALT_T', alt_threshold)
                    for index, sample in enumerate(record.samples):
                        ref_count = ref_filtered_count_array[index]
                        alt_count = alt_filtered_count_array[index]
                        new_record.samples[sample]['AD'] = (ref_count, alt_count)
                        genotype = self.get_genotype(ref_count, alt_count)
                        new_record.samples[sample]['GT'] = genotype

                    vcf_out.write(new_record)


def filter_snp(args):
    with Filter_snp(args) as runner:
        runner.run()


def get_opts_filter_snp(parser, sub_program):
    parser.add_argument('--ref_threshold_method', default='otsu', choices=['otsu', 'auto', 'none'], help=HELP_DICT['threshold_method'])
    parser.add_argument('--alt_threshold_method', default='otsu', choices=['otsu', 'auto', 'none'], help=HELP_DICT['threshold_method'])
    parser.add_argument(
        "--ref_min_support_read",
        type=int,
        help='minimum supporting read number for ref.', 
        default=2,
    )
    parser.add_argument(
        "--alt_min_support_read",
        type=int,
        help='minimum supporting read number for alt.', 
        default=2,
    )
    if sub_program:
        parser.add_argument("--vcf", help="norm vcf file")
        s_common(parser)




        





